# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for acme.utils.reverb_utils."""

from absl.testing import absltest
from acme import types
from acme.adders import reverb as reverb_adders
from acme.utils import reverb_utils
import numpy as np
import reverb
import tree


class ReverbUtilsTest(absltest.TestCase):

  def test_make_replay_table_preserves_table_info(self):
    limiter = reverb.rate_limiters.SampleToInsertRatio(
        samples_per_insert=1, min_size_to_sample=2, error_buffer=(0, 10))
    table = reverb.Table(
        name='test',
        sampler=reverb.selectors.Uniform(),
        remover=reverb.selectors.Fifo(),
        max_size=10,
        rate_limiter=limiter)
    new_table = reverb_utils.make_replay_table_from_info(table.info)
    new_info = new_table.info

    # table_worker_time is not set by the above utility since this is meant to
    # be monitoring information about any given table. So instead we copy this
    # so that the assertion below checks that everything else matches.

    new_info.table_worker_time.sleeping_ms = (
        table.info.table_worker_time.sleeping_ms)

    self.assertEqual(new_info, table.info)

  _EMPTY_INFO = reverb.SampleInfo((), (), (), ())
  _DUMMY_OBS = np.array([[[0], [1], [2]]])
  _DUMMY_ACTION = np.array([[[3], [4], [5]]])
  _DUMMY_REWARD = np.array([[6, 7, 8]])
  _DUMMY_DISCOUNT = np.array([[.99, .99, .99]])
  _DUMMY_NEXT_OBS = np.array([[[1], [2], [0]]])

  def _create_dummy_steps(self):
    return reverb_adders.Step(
        observation=self._DUMMY_OBS,
        action=self._DUMMY_ACTION,
        reward=self._DUMMY_REWARD,
        discount=self._DUMMY_DISCOUNT,
        start_of_episode=True,
        extras=())

  def _create_dummy_transitions(self):
    return types.Transition(
        observation=self._DUMMY_OBS,
        action=self._DUMMY_ACTION,
        reward=self._DUMMY_REWARD,
        discount=self._DUMMY_DISCOUNT,
        next_observation=self._DUMMY_NEXT_OBS)

  def test_replay_sample_to_sars_transition_is_sequence(self):
    fake_sample = reverb.ReplaySample(
        info=self._EMPTY_INFO, data=self._create_dummy_steps())
    fake_transition = self._create_dummy_transitions()
    transition_from_sample = reverb_utils.replay_sample_to_sars_transition(
        fake_sample, is_sequence=True)
    tree.map_structure(np.testing.assert_array_equal, transition_from_sample,
                       fake_transition)


if __name__ == '__main__':
  absltest.main()



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for Episode adders."""

from absl.testing import absltest
from absl.testing import parameterized

from acme.adders.reverb import episode as adders
from acme.adders.reverb import test_utils
import numpy as np


class EpisodeAdderTest(test_utils.AdderTestMixin, parameterized.TestCase):

  @parameterized.parameters(2, 10, 50)
  def test_adder(self, max_sequence_length):
    adder = adders.EpisodeAdder(self.client, max_sequence_length)

    # Create a simple trajectory to add.
    observations = range(max_sequence_length)
    first, steps = test_utils.make_trajectory(observations)

    expected_episode = test_utils.make_sequence(observations)
    self.run_test_adder(
        adder=adder,
        first=first,
        steps=steps,
        expected_items=[expected_episode],
        signature=adder.signature(*test_utils.get_specs(steps[0])))

  @parameterized.parameters(2, 10, 50)
  def test_max_sequence_length(self, max_sequence_length):
    adder = adders.EpisodeAdder(self.client, max_sequence_length)

    first, steps = test_utils.make_trajectory(range(max_sequence_length + 1))
    adder.add_first(first)
    for action, step in steps[:-1]:
      adder.add(action, step)

    # We should have max_sequence_length-1 timesteps that have been written,
    # where the -1 is due to the dangling observation (ie we have actually
    # seen max_sequence_length observations).
    self.assertEqual(self.num_items(), 0)

    # Adding one more step should raise an error.
    with self.assertRaises(ValueError):
      action, step = steps[-1]
      adder.add(action, step)

    # Since the last insert failed it should not affect the internal state.
    self.assertEqual(self.num_items(), 0)

  @parameterized.parameters((2, 1), (10, 2), (50, 5))
  def test_padding(self, max_sequence_length, padding):
    adder = adders.EpisodeAdder(
        self.client,
        max_sequence_length + padding,
        padding_fn=np.zeros)

    # Create a simple trajectory to add.
    observations = range(max_sequence_length)
    first, steps = test_utils.make_trajectory(observations)

    expected_episode = test_utils.make_sequence(observations)
    for _ in range(padding):
      expected_episode.append((0, 0, 0.0, 0.0, False, ()))

    self.run_test_adder(
        adder=adder,
        first=first,
        steps=steps,
        expected_items=[expected_episode],
        signature=adder.signature(*test_utils.get_specs(steps[0])))

  @parameterized.parameters((2, 1), (10, 2), (50, 5))
  def test_nonzero_padding(self, max_sequence_length, padding):
    adder = adders.EpisodeAdder(
        self.client,
        max_sequence_length + padding,
        padding_fn=lambda s, d: np.zeros(s, d) - 1)

    # Create a simple trajectory to add.
    observations = range(max_sequence_length)
    first, steps = test_utils.make_trajectory(observations)

    expected_episode = test_utils.make_sequence(observations)
    for _ in range(padding):
      expected_episode.append((-1, -1, -1.0, -1.0, False, ()))

    self.run_test_adder(
        adder=adder,
        first=first,
        steps=steps,
        expected_items=[expected_episode],
        signature=adder.signature(*test_utils.get_specs(steps[0])))


if __name__ == '__main__':
  absltest.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Learner for the PPO agent."""

from typing import Dict, Iterator, List, NamedTuple, Optional, Tuple

import acme
from acme import types
from acme.agents.jax.ppo import networks
from acme.jax import networks as networks_lib
from acme.jax import utils
from acme.utils import counting
from acme.utils import loggers
import jax
import jax.numpy as jnp
import numpy as np
import optax
import reverb
import rlax


class Batch(NamedTuple):
  """A batch of data; all shapes are expected to be [B, ...]."""
  observations: types.NestedArray
  actions: jnp.ndarray
  advantages: jnp.ndarray

  # Target value estimate used to bootstrap the value function.
  target_values: jnp.ndarray

  # Value estimate and action log-prob at behavior time.
  behavior_values: jnp.ndarray
  behavior_log_probs: jnp.ndarray


class TrainingState(NamedTuple):
  """Training state consists of network parameters and optimiser state."""
  params: networks_lib.Params
  opt_state: optax.OptState
  random_key: networks_lib.PRNGKey


class PPOLearner(acme.Learner):
  """Learner for PPO."""

  def __init__(
      self,
      ppo_networks: networks.PPONetworks,
      iterator: Iterator[reverb.ReplaySample],
      optimizer: optax.GradientTransformation,
      random_key: networks_lib.PRNGKey,
      ppo_clipping_epsilon: float = 0.2,
      clip_value: bool = True,
      gae_lambda: float = 0.95,
      discount: float = 0.99,
      entropy_cost: float = 0.,
      value_cost: float = 1.,
      max_abs_reward: float = np.inf,
      num_epochs: int = 4,
      num_minibatches: int = 1,
      counter: Optional[counting.Counter] = None,
      logger: Optional[loggers.Logger] = None,
  ):

    def gae_advantages(rewards: jnp.array, discounts: jnp.array,
                       values: jnp.array) -> Tuple[jnp.ndarray, jnp.array]:
      """Uses truncated GAE to compute advantages."""

      # Apply reward clipping.
      rewards = jnp.clip(rewards, -max_abs_reward, max_abs_reward)

      advantages = rlax.truncated_generalized_advantage_estimation(
          rewards[:-1], discounts[:-1], gae_lambda, values)
      advantages = jax.lax.stop_gradient(advantages)

      # Exclude the bootstrap value
      target_values = values[:-1] + advantages
      target_values = jax.lax.stop_gradient(target_values)

      return advantages, target_values

    def loss(
        params: networks_lib.Params, observations: types.NestedArray,
        actions: jnp.array, behaviour_log_probs: jnp.array,
        target_values: jnp.array, advantages: jnp.array,
        behavior_values: jnp.array
    ) -> Tuple[jnp.ndarray, Dict[str, jnp.ndarray]]:
      """Surrogate loss using clipped probability ratios."""

      distribution_params, values = ppo_networks.network.apply(params,
                                                               observations)
      log_probs = ppo_networks.log_prob(distribution_params, actions)
      entropy = ppo_networks.entropy(distribution_params)

      # Compute importance sampling weights: current policy / behavior policy.
      rhos = jnp.exp(log_probs - behaviour_log_probs)

      policy_loss = rlax.clipped_surrogate_pg_loss(rhos, advantages,
                                                   ppo_clipping_epsilon)

      # Value function loss. Exclude the bootstrap value
      unclipped_value_error = target_values - values
      unclipped_value_loss = unclipped_value_error ** 2

      if clip_value:
        # Clip values to reduce variablility during critic training.
        clipped_values = behavior_values + jnp.clip(
            values - behavior_values, -ppo_clipping_epsilon,
            ppo_clipping_epsilon)
        clipped_value_error = target_values - clipped_values
        clipped_value_loss = clipped_value_error ** 2
        value_loss = jnp.mean(jnp.fmax(unclipped_value_loss,
                                       clipped_value_loss))
      else:
        # For Mujoco envs clipping hurts a lot. Evidenced by Figure 43 in
        # https://arxiv.org/pdf/2006.05990.pdf
        value_loss = jnp.mean(unclipped_value_loss)

      # Entropy regulariser.
      entropy_loss = -jnp.mean(entropy)

      total_loss = (
          policy_loss + value_loss * value_cost + entropy_loss * entropy_cost)
      return total_loss, {
          'loss_total': total_loss,
          'loss_policy': policy_loss,
          'loss_value': value_loss,
          'loss_entropy': entropy_loss,
      }

    @jax.jit
    def sgd_step(
        state: TrainingState, sample: reverb.ReplaySample
    ) -> Tuple[TrainingState, Dict[str, jnp.ndarray]]:
      """Performs a minibatch SGD step, returning new state and metrics."""

      # Extract the data.
      data = sample.data
      # TODO(sinopalnikov): replace it with namedtuple unpacking
      observations, actions, rewards, termination, extra = (data.observation,
                                                            data.action,
                                                            data.reward,
                                                            data.discount,
                                                            data.extras)
      discounts = termination * discount
      behavior_log_probs = extra['log_prob']

      def get_behavior_values(params: networks_lib.Params,
                              observations: types.NestedArray) -> jnp.ndarray:
        o = jax.tree_map(lambda x: jnp.reshape(x, [-1] + list(x.shape[2:])),
                         observations)
        _, behavior_values = ppo_networks.network.apply(params, o)
        behavior_values = jnp.reshape(behavior_values, rewards.shape[0:2])
        return behavior_values

      behavior_values = get_behavior_values(state.params, observations)

      # Vmap over batch dimension
      batch_gae_advantages = jax.vmap(gae_advantages, in_axes=0)
      advantages, target_values = batch_gae_advantages(rewards, discounts,
                                                       behavior_values)
      # Exclude the last step - it was only used for bootstrapping.
      # The shape is [num_sequences, num_steps, ..]
      observations, actions, behavior_log_probs, behavior_values = jax.tree_map(
          lambda x: x[:, :-1],
          (observations, actions, behavior_log_probs, behavior_values))
      trajectories = Batch(observations=observations,
                           actions=actions,
                           advantages=advantages,
                           behavior_log_probs=behavior_log_probs,
                           target_values=target_values,
                           behavior_values=behavior_values)

      # Concatenate all trajectories. Reshape from [num_sequences, num_steps,..]
      # to [num_sequences * num_steps,..]
      assert len(target_values.shape) > 1
      num_sequences = target_values.shape[0]
      num_steps = target_values.shape[1]
      batch_size = num_sequences * num_steps
      assert batch_size % num_minibatches == 0, (
          'Num minibatches must divide batch size. Got batch_size={}'
          ' num_minibatches={}.'
      ).format(batch_size, num_minibatches)
      batch = jax.tree_map(lambda x: x.reshape((batch_size,) + x.shape[2:]),
                           trajectories)

      # Compute gradients.
      grad_fn = jax.grad(loss, has_aux=True)

      def model_update_minibatch(
          carry: Tuple[networks_lib.Params, optax.OptState],
          minibatch: Batch,
      ) -> Tuple[Tuple[networks_lib.Params, optax.OptState], Dict[
          str, jnp.ndarray]]:
        """Performs model update for a single minibatch."""
        params, opt_state = carry
        # Normalize advantages at the minibatch level before using them.
        advantages = ((minibatch.advantages -
                       jnp.mean(minibatch.advantages, axis=0)) /
                      (jnp.std(minibatch.advantages, axis=0) + 1e-8))
        gradients, metrics = grad_fn(params,
                                     minibatch.observations,
                                     minibatch.actions,
                                     minibatch.behavior_log_probs,
                                     minibatch.target_values,
                                     advantages,
                                     minibatch.behavior_values)

        # Apply updates
        updates, opt_state = optimizer.update(gradients, opt_state)
        params = optax.apply_updates(params, updates)

        metrics['norm_grad'] = optax.global_norm(gradients)
        metrics['norm_updates'] = optax.global_norm(updates)
        return (params, opt_state), metrics

      def model_update_epoch(
          carry: Tuple[jnp.ndarray, networks_lib.Params, optax.OptState,
                       Batch], unused_t: Tuple[()]
      ) -> Tuple[Tuple[jnp.ndarray, networks_lib.Params, optax.OptState,
                       Batch], Dict[str, jnp.ndarray]]:
        """Performs model updates based on one epoch of data."""
        key, params, opt_state, batch = carry
        key, subkey = jax.random.split(key)
        permutation = jax.random.permutation(subkey, batch_size)
        shuffled_batch = jax.tree_map(
            lambda x: jnp.take(x, permutation, axis=0), batch)
        minibatches = jax.tree_map(
            lambda x: jnp.reshape(x, [num_minibatches, -1] + list(x.shape[1:])),
            shuffled_batch)

        (params, opt_state), metrics = jax.lax.scan(
            model_update_minibatch, (params, opt_state), minibatches,
            length=num_minibatches)

        return (key, params, opt_state, batch), metrics

      params = state.params
      opt_state = state.opt_state
      # Repeat training for the given number of epoch, taking a random
      # permutation for every epoch.
      (key, params, opt_state, _), metrics = jax.lax.scan(
          model_update_epoch, (state.random_key, params, opt_state, batch), (),
          length=num_epochs)

      metrics = jax.tree_map(jnp.mean, metrics)
      metrics['norm_params'] = optax.global_norm(params)
      metrics['observations_mean'] = jnp.mean(
          utils.batch_concat(
              jax.tree_map(lambda x: jnp.abs(jnp.mean(x, axis=(0, 1))),
                           observations),
              num_batch_dims=0))
      metrics['observations_std'] = jnp.mean(
          utils.batch_concat(
              jax.tree_map(lambda x: jnp.std(x, axis=(0, 1)), observations),
              num_batch_dims=0))
      metrics['rewards_mean'] = jnp.mean(
          jnp.abs(jnp.mean(rewards, axis=(0, 1))))
      metrics['rewards_std'] = jnp.std(rewards, axis=(0, 1))
      new_state = TrainingState(params=params,
                                opt_state=opt_state,
                                random_key=key)
      return new_state, metrics

    def make_initial_state(key: networks_lib.PRNGKey) -> TrainingState:
      """Initialises the training state (parameters and optimiser state)."""
      key_init, key_state = jax.random.split(key)
      initial_params = ppo_networks.network.init(key_init)
      initial_opt_state = optimizer.init(initial_params)
      return TrainingState(
          params=initial_params,
          opt_state=initial_opt_state,
          random_key=key_state)

    # Initialise training state (parameters and optimiser state).
    self._state = make_initial_state(random_key)

    # Internalise iterator.
    self._iterator = iterator
    self._sgd_step = sgd_step

    # Set up logging/counting.
    self._counter = counter or counting.Counter()
    self._logger = logger or loggers.make_default_logger('learner')

  def step(self):
    """Does a step of SGD and logs the results."""

    # Do a batch of SGD.
    sample = next(self._iterator)
    self._state, results = self._sgd_step(self._state, sample)

    # Update our counts and record it.
    counts = self._counter.increment(steps=1)

    # Snapshot and attempt to write logs.
    self._logger.write({**results, **counts})

  def get_variables(self, names: List[str]) -> List[networks_lib.Params]:
    return [self._state.params]

  def save(self) -> TrainingState:
    return self._state

  def restore(self, state: TrainingState):
    self._state = state



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""D4PG learner implementation."""

import time
from typing import Dict, Iterator, List, NamedTuple, Optional, Tuple

import acme
from acme import types
from acme.jax import networks as networks_lib
from acme.jax import utils
from acme.utils import counting
from acme.utils import loggers
import jax
import jax.numpy as jnp
import optax
import reverb
import rlax


class TrainingState(NamedTuple):
  """Contains training state for the learner."""
  policy_params: networks_lib.Params
  target_policy_params: networks_lib.Params
  critic_params: networks_lib.Params
  target_critic_params: networks_lib.Params
  policy_opt_state: optax.OptState
  critic_opt_state: optax.OptState
  steps: int


class D4PGLearner(acme.Learner):
  """D4PG learner.

  This is the learning component of a D4PG agent. IE it takes a dataset as input
  and implements update functionality to learn from this dataset.
  """

  _state: TrainingState

  def __init__(self,
               policy_network: networks_lib.FeedForwardNetwork,
               critic_network: networks_lib.FeedForwardNetwork,
               random_key: networks_lib.PRNGKey,
               discount: float,
               target_update_period: int,
               iterator: Iterator[reverb.ReplaySample],
               policy_optimizer: Optional[optax.GradientTransformation] = None,
               critic_optimizer: Optional[optax.GradientTransformation] = None,
               clipping: bool = True,
               counter: Optional[counting.Counter] = None,
               logger: Optional[loggers.Logger] = None,
               jit: bool = True,
               num_sgd_steps_per_step: int = 1):

    def critic_mean(
        critic_params: networks_lib.Params,
        observation: types.NestedArray,
        action: types.NestedArray,
    ) -> jnp.ndarray:
      # We add batch dimension to make sure batch concat in critic_network
      # works correctly.
      observation = utils.add_batch_dim(observation)
      action = utils.add_batch_dim(action)
      # Computes the mean action-value estimate.
      logits, atoms = critic_network.apply(critic_params, observation, action)
      logits = utils.squeeze_batch_dim(logits)
      probabilities = jax.nn.softmax(logits)
      return jnp.sum(probabilities * atoms, axis=-1)

    def policy_loss(
        policy_params: networks_lib.Params,
        critic_params: networks_lib.Params,
        o_t: types.NestedArray,
    ) -> jnp.ndarray:
      # Computes the discrete policy gradient loss.
      dpg_a_t = policy_network.apply(policy_params, o_t)
      grad_critic = jax.vmap(
          jax.grad(critic_mean, argnums=2), in_axes=(None, 0, 0))
      dq_da = grad_critic(critic_params, o_t, dpg_a_t)
      dqda_clipping = 1. if clipping else None
      batch_dpg_learning = jax.vmap(rlax.dpg_loss, in_axes=(0, 0, None))
      loss = batch_dpg_learning(dpg_a_t, dq_da, dqda_clipping)
      return jnp.mean(loss)

    def critic_loss(
        critic_params: networks_lib.Params,
        state: TrainingState,
        transition: types.Transition,
    ):
      # Computes the distributional critic loss.
      q_tm1, atoms_tm1 = critic_network.apply(critic_params,
                                              transition.observation,
                                              transition.action)
      a = policy_network.apply(state.target_policy_params,
                               transition.next_observation)
      q_t, atoms_t = critic_network.apply(state.target_critic_params,
                                          transition.next_observation, a)
      batch_td_learning = jax.vmap(
          rlax.categorical_td_learning, in_axes=(None, 0, 0, 0, None, 0))
      loss = batch_td_learning(atoms_tm1, q_tm1, transition.reward,
                               discount * transition.discount, atoms_t, q_t)
      return jnp.mean(loss)

    def sgd_step(
        state: TrainingState,
        transitions: types.Transition,
    ) -> Tuple[TrainingState, Dict[str, jnp.ndarray]]:

      # TODO(jaslanides): Use a shared forward pass for efficiency.
      policy_loss_and_grad = jax.value_and_grad(policy_loss)
      critic_loss_and_grad = jax.value_and_grad(critic_loss)

      # Compute losses and their gradients.
      policy_loss_value, policy_gradients = policy_loss_and_grad(
          state.policy_params, state.critic_params,
          transitions.next_observation)
      critic_loss_value, critic_gradients = critic_loss_and_grad(
          state.critic_params, state, transitions)

      # Get optimizer updates and state.
      policy_updates, policy_opt_state = policy_optimizer.update(  # pytype: disable=attribute-error
          policy_gradients, state.policy_opt_state)
      critic_updates, critic_opt_state = critic_optimizer.update(  # pytype: disable=attribute-error
          critic_gradients, state.critic_opt_state)

      # Apply optimizer updates to parameters.
      policy_params = optax.apply_updates(state.policy_params, policy_updates)
      critic_params = optax.apply_updates(state.critic_params, critic_updates)

      steps = state.steps + 1

      # Periodically update target networks.
      target_policy_params, target_critic_params = rlax.periodic_update(
          (policy_params, critic_params),
          (state.target_policy_params, state.target_critic_params),
          steps, self._target_update_period)

      new_state = TrainingState(
          policy_params=policy_params,
          critic_params=critic_params,
          target_policy_params=target_policy_params,
          target_critic_params=target_critic_params,
          policy_opt_state=policy_opt_state,
          critic_opt_state=critic_opt_state,
          steps=steps,
      )

      metrics = {
          'policy_loss': policy_loss_value,
          'critic_loss': critic_loss_value,
      }

      return new_state, metrics

    # General learner book-keeping and loggers.
    self._counter = counter or counting.Counter()
    self._logger = logger or loggers.make_default_logger(
        'learner', asynchronous=True, serialize_fn=utils.fetch_devicearray)

    # Necessary to track when to update target networks.
    self._target_update_period = target_update_period

    # Create prefetching dataset iterator.
    self._iterator = iterator

    # Maybe use the JIT compiler.
    sgd_step = utils.process_multiple_batches(sgd_step, num_sgd_steps_per_step)
    self._sgd_step = jax.jit(sgd_step) if jit else sgd_step

    # Create the network parameters and copy into the target network parameters.
    key_policy, key_critic = jax.random.split(random_key)
    initial_policy_params = policy_network.init(key_policy)
    initial_critic_params = critic_network.init(key_critic)
    initial_target_policy_params = initial_policy_params
    initial_target_critic_params = initial_critic_params

    # Create optimizers if they aren't given.
    critic_optimizer = critic_optimizer or optax.adam(1e-4)
    policy_optimizer = policy_optimizer or optax.adam(1e-4)

    # Initialize optimizers.
    initial_policy_opt_state = policy_optimizer.init(initial_policy_params)  # pytype: disable=attribute-error
    initial_critic_opt_state = critic_optimizer.init(initial_critic_params)  # pytype: disable=attribute-error

    # Create initial state.
    self._state = TrainingState(
        policy_params=initial_policy_params,
        target_policy_params=initial_target_policy_params,
        critic_params=initial_critic_params,
        target_critic_params=initial_target_critic_params,
        policy_opt_state=initial_policy_opt_state,
        critic_opt_state=initial_critic_opt_state,
        steps=0,
    )

    # Do not record timestamps until after the first learning step is done.
    # This is to avoid including the time it takes for actors to come online and
    # fill the replay buffer.
    self._timestamp = None

  def step(self):
    # Get data from replay (dropping extras if any). Note there is no
    # extra data here because we do not insert any into Reverb.
    sample = next(self._iterator)
    transitions = types.Transition(*sample.data)

    self._state, metrics = self._sgd_step(self._state, transitions)

    # Compute elapsed time.
    timestamp = time.time()
    elapsed_time = timestamp - self._timestamp if self._timestamp else 0
    self._timestamp = timestamp

    # Increment counts and record the current time
    counts = self._counter.increment(steps=1, walltime=elapsed_time)

    # Attempts to write the logs.
    self._logger.write({**metrics, **counts})

  def get_variables(self, names: List[str]) -> List[networks_lib.Params]:
    variables = {
        'policy': self._state.target_policy_params,
        'critic': self._state.target_critic_params,
    }
    return [variables[name] for name in names]

  def save(self) -> TrainingState:
    return self._state

  def restore(self, state: TrainingState):
    self._state = state



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the noop starts wrapper."""

from unittest import mock

from absl.testing import absltest
from acme import wrappers
from acme.testing import fakes
from dm_env import specs
import numpy as np


class NoopStartsTest(absltest.TestCase):

  def test_reset(self):
    """Ensure that noop starts `reset` steps the environment multiple times."""
    noop_action = 0
    noop_max = 10
    seed = 24

    base_env = fakes.DiscreteEnvironment(
        action_dtype=np.int64,
        obs_dtype=np.int64,
        reward_spec=specs.Array(dtype=np.float64, shape=()))
    mock_step_fn = mock.MagicMock()
    expected_num_step_calls = np.random.RandomState(seed).randint(noop_max + 1)

    with mock.patch.object(base_env, 'step', mock_step_fn):
      env = wrappers.NoopStartsWrapper(
          base_env,
          noop_action=noop_action,
          noop_max=noop_max,
          seed=seed,
      )
      env.reset()

      # Test environment step called with noop action as part of wrapper.reset
      mock_step_fn.assert_called_with(noop_action)
      self.assertEqual(mock_step_fn.call_count, expected_num_step_calls)
      self.assertEqual(mock_step_fn.call_args, ((noop_action,), {}))

  def test_raises_value_error(self):
    """Ensure that wrapper raises error if noop_max is <0."""
    base_env = fakes.DiscreteEnvironment(
        action_dtype=np.int64,
        obs_dtype=np.int64,
        reward_spec=specs.Array(dtype=np.float64, shape=()))

    with self.assertRaises(ValueError):
      wrappers.NoopStartsWrapper(base_env, noop_action=0, noop_max=-1, seed=24)


if __name__ == '__main__':
  absltest.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements the multi-objective MPO (MO-MPO) loss.

This loss was proposed in (Abdolmaleki, Huang et al., 2020).

The loss is implemented as a Sonnet module rather than a function so that it
can hold its own dual variables, as instances of `tf.Variable`, which it creates
the first time the module is called.

Tensor shapes are annotated, where helpful, as follow:
  B: batch size,
  N: number of sampled actions, see MO-MPO paper for more details,
  D: dimensionality of the action space,
  K: number of objectives.

(Abdolmaleki, Huang et al., 2020): https://arxiv.org/pdf/2005.07513.pdf
"""

import dataclasses
from typing import Dict, Sequence, Tuple, Union

from acme.tf.losses import mpo
import sonnet as snt
import tensorflow as tf
import tensorflow_probability as tfp

tfd = tfp.distributions

_MPO_FLOAT_EPSILON = 1e-8


@dataclasses.dataclass
class KLConstraint:
  """Defines a per-objective policy improvement step constraint for MO-MPO."""

  name: str
  value: float

  def __post_init__(self):
    if self.value < 0:
      raise ValueError("KL constraint epsilon must be non-negative.")


class MultiObjectiveMPO(snt.Module):
  """Multi-objective MPO loss with decoupled KL constraints.

  This implementation of the MO-MPO loss is based on the approach proposed in
  (Abdolmaleki, Huang et al., 2020). The following features are included as
  options:
  - Satisfying the KL-constraint on a per-dimension basis (on by default)

  (Abdolmaleki, Huang et al., 2020): https://arxiv.org/pdf/2005.07513.pdf
  """

  def __init__(self,
               epsilons: Sequence[KLConstraint],
               epsilon_mean: float,
               epsilon_stddev: float,
               init_log_temperature: float,
               init_log_alpha_mean: float,
               init_log_alpha_stddev: float,
               per_dim_constraining: bool = True,
               name: str = "MOMPO"):
    """Initialize and configure the MPO loss.

    Args:
      epsilons: per-objective KL constraints on the non-parametric auxiliary
        policy, the one associated with the dual variables called temperature;
        expected length K.
      epsilon_mean: KL constraint on the mean of the Gaussian policy, the one
        associated with the dual variable called alpha_mean.
      epsilon_stddev: KL constraint on the stddev of the Gaussian policy, the
        one associated with the dual variable called alpha_mean.
      init_log_temperature: initial value for the temperature in log-space, note
        a softplus (rather than an exp) will be used to transform this.
      init_log_alpha_mean: initial value for the alpha_mean in log-space, note
        a softplus (rather than an exp) will be used to transform this.
      init_log_alpha_stddev: initial value for the alpha_stddev in log-space,
        note a softplus (rather than an exp) will be used to transform this.
      per_dim_constraining: whether to enforce the KL constraint on each
        dimension independently; this is the default. Otherwise the overall KL
        is constrained, which allows some dimensions to change more at the
        expense of others staying put.
      name: a name for the module, passed directly to snt.Module.

    """
    super().__init__(name=name)

    # MO-MPO constraint thresholds.
    self._epsilons = tf.constant([x.value for x in epsilons])
    self._epsilon_mean = tf.constant(epsilon_mean)
    self._epsilon_stddev = tf.constant(epsilon_stddev)

    # Initial values for the constraints' dual variables.
    self._init_log_temperature = init_log_temperature
    self._init_log_alpha_mean = init_log_alpha_mean
    self._init_log_alpha_stddev = init_log_alpha_stddev

    # Whether to ensure per-dimension KL constraint satisfication.
    self._per_dim_constraining = per_dim_constraining

    # Remember the number of objectives
    self._num_objectives = len(epsilons)  # K = number of objectives
    self._objective_names = [x.name for x in epsilons]

    # Make sure there are no duplicate objective names
    if len(self._objective_names) != len(set(self._objective_names)):
      raise ValueError("Duplicate objective names are not allowed.")

  @property
  def objective_names(self):
    return self._objective_names

  @snt.once
  def create_dual_variables_once(self, shape: tf.TensorShape, dtype: tf.DType):
    """Creates the dual variables the first time the loss module is called."""

    # Create the dual variables.
    self._log_temperature = tf.Variable(
        initial_value=[self._init_log_temperature] * self._num_objectives,
        dtype=dtype,
        name="log_temperature",
        shape=(self._num_objectives,))
    self._log_alpha_mean = tf.Variable(
        initial_value=tf.fill(shape, self._init_log_alpha_mean),
        dtype=dtype,
        name="log_alpha_mean",
        shape=shape)
    self._log_alpha_stddev = tf.Variable(
        initial_value=tf.fill(shape, self._init_log_alpha_stddev),
        dtype=dtype,
        name="log_alpha_stddev",
        shape=shape)

    # Cast constraint thresholds to the expected dtype.
    self._epsilons = tf.cast(self._epsilons, dtype)
    self._epsilon_mean = tf.cast(self._epsilon_mean, dtype)
    self._epsilon_stddev = tf.cast(self._epsilon_stddev, dtype)

  def __call__(
      self,
      online_action_distribution: Union[tfd.MultivariateNormalDiag,
                                        tfd.Independent],
      target_action_distribution: Union[tfd.MultivariateNormalDiag,
                                        tfd.Independent],
      actions: tf.Tensor,  # Shape [N, B, D].
      q_values: tf.Tensor,  # Shape [N, B, K].
  ) -> Tuple[tf.Tensor, Dict[str, tf.Tensor]]:
    """Computes the decoupled MO-MPO loss.

    Args:
      online_action_distribution: online distribution returned by the online
        policy network; expects batch_dims of [B] and event_dims of [D].
      target_action_distribution: target distribution returned by the target
        policy network; expects same shapes as online distribution.
      actions: actions sampled from the target policy; expects shape [N, B, D].
      q_values: Q-values associated with each action; expects shape [N, B, K].

    Returns:
      Loss, combining the policy loss, KL penalty, and dual losses required to
        adapt the dual variables.
      Stats, for diagnostics and tracking performance.
    """

    # Make sure the Q-values are per-objective
    q_values.get_shape().assert_has_rank(3)
    if q_values.get_shape()[-1] != self._num_objectives:
      raise ValueError("Q-values do not match expected number of objectives.")

    # Cast `MultivariateNormalDiag`s to Independent Normals.
    # The latter allows us to satisfy KL constraints per-dimension.
    if isinstance(target_action_distribution, tfd.MultivariateNormalDiag):
      target_action_distribution = tfd.Independent(
          tfd.Normal(target_action_distribution.mean(),
                     target_action_distribution.stddev()))
      online_action_distribution = tfd.Independent(
          tfd.Normal(online_action_distribution.mean(),
                     online_action_distribution.stddev()))

    # Infer the shape and dtype of dual variables.
    scalar_dtype = q_values.dtype
    if self._per_dim_constraining:
      dual_variable_shape = target_action_distribution.distribution.kl_divergence(
          online_action_distribution.distribution).shape[1:]  # Should be [D].
    else:
      dual_variable_shape = target_action_distribution.kl_divergence(
          online_action_distribution).shape[1:]  # Should be [1].

    # Create dual variables for the KL constraints; only happens the first call.
    self.create_dual_variables_once(dual_variable_shape, scalar_dtype)

    # Project dual variables to ensure they stay positive.
    min_log_temperature = tf.constant(-18.0, scalar_dtype)
    min_log_alpha = tf.constant(-18.0, scalar_dtype)
    self._log_temperature.assign(
        tf.maximum(min_log_temperature, self._log_temperature))
    self._log_alpha_mean.assign(tf.maximum(min_log_alpha, self._log_alpha_mean))
    self._log_alpha_stddev.assign(
        tf.maximum(min_log_alpha, self._log_alpha_stddev))

    # Transform dual variables from log-space.
    # Note: using softplus instead of exponential for numerical stability.
    temperature = tf.math.softplus(self._log_temperature) + _MPO_FLOAT_EPSILON
    alpha_mean = tf.math.softplus(self._log_alpha_mean) + _MPO_FLOAT_EPSILON
    alpha_stddev = tf.math.softplus(self._log_alpha_stddev) + _MPO_FLOAT_EPSILON

    # Get online and target means and stddevs in preparation for decomposition.
    online_mean = online_action_distribution.distribution.mean()
    online_scale = online_action_distribution.distribution.stddev()
    target_mean = target_action_distribution.distribution.mean()
    target_scale = target_action_distribution.distribution.stddev()

    # Compute normalized importance weights, used to compute expectations with
    # respect to the non-parametric policy; and the temperature loss, used to
    # adapt the tempering of Q-values.
    normalized_weights, loss_temperature = compute_weights_and_temperature_loss(
        q_values, self._epsilons, temperature)  # Shapes [N, B, K] and [1, K].
    normalized_weights_sum = tf.reduce_sum(normalized_weights, axis=-1)
    loss_temperature_mean = tf.reduce_mean(loss_temperature)

    # Only needed for diagnostics: Compute estimated actualized KL between the
    # non-parametric and current target policies.
    kl_nonparametric = mpo.compute_nonparametric_kl_from_normalized_weights(
        normalized_weights)

    # Decompose the online policy into fixed-mean & fixed-stddev distributions.
    # This has been documented as having better performance in bandit settings,
    # see e.g. https://arxiv.org/pdf/1812.02256.pdf.
    fixed_stddev_distribution = tfd.Independent(
        tfd.Normal(loc=online_mean, scale=target_scale))
    fixed_mean_distribution = tfd.Independent(
        tfd.Normal(loc=target_mean, scale=online_scale))

    # Compute the decomposed policy losses.
    loss_policy_mean = mpo.compute_cross_entropy_loss(
        actions, normalized_weights_sum, fixed_stddev_distribution)
    loss_policy_stddev = mpo.compute_cross_entropy_loss(
        actions, normalized_weights_sum, fixed_mean_distribution)

    # Compute the decomposed KL between the target and online policies.
    if self._per_dim_constraining:
      kl_mean = target_action_distribution.distribution.kl_divergence(
          fixed_stddev_distribution.distribution)  # Shape [B, D].
      kl_stddev = target_action_distribution.distribution.kl_divergence(
          fixed_mean_distribution.distribution)  # Shape [B, D].
    else:
      kl_mean = target_action_distribution.kl_divergence(
          fixed_stddev_distribution)  # Shape [B].
      kl_stddev = target_action_distribution.kl_divergence(
          fixed_mean_distribution)  # Shape [B].

    # Compute the alpha-weighted KL-penalty and dual losses to adapt the alphas.
    loss_kl_mean, loss_alpha_mean = mpo.compute_parametric_kl_penalty_and_dual_loss(
        kl_mean, alpha_mean, self._epsilon_mean)
    loss_kl_stddev, loss_alpha_stddev = mpo.compute_parametric_kl_penalty_and_dual_loss(
        kl_stddev, alpha_stddev, self._epsilon_stddev)

    # Combine losses.
    loss_policy = loss_policy_mean + loss_policy_stddev
    loss_kl_penalty = loss_kl_mean + loss_kl_stddev
    loss_dual = loss_alpha_mean + loss_alpha_stddev + loss_temperature_mean
    loss = loss_policy + loss_kl_penalty + loss_dual

    stats = {}
    # Dual Variables.
    stats["dual_alpha_mean"] = tf.reduce_mean(alpha_mean)
    stats["dual_alpha_stddev"] = tf.reduce_mean(alpha_stddev)
    # Losses.
    stats["loss_policy"] = tf.reduce_mean(loss)
    stats["loss_alpha"] = tf.reduce_mean(loss_alpha_mean + loss_alpha_stddev)
    # KL measurements.
    stats["kl_mean_rel"] = tf.reduce_mean(kl_mean, axis=0) / self._epsilon_mean
    stats["kl_stddev_rel"] = tf.reduce_mean(
        kl_stddev, axis=0) / self._epsilon_stddev
    # If the policy has standard deviation, log summary stats for this as well.
    pi_stddev = online_action_distribution.distribution.stddev()
    stats["pi_stddev_min"] = tf.reduce_mean(tf.reduce_min(pi_stddev, axis=-1))
    stats["pi_stddev_max"] = tf.reduce_mean(tf.reduce_max(pi_stddev, axis=-1))

    # Condition number of the diagonal covariance (actually, stddev) matrix.
    stats["pi_stddev_cond"] = tf.reduce_mean(
        tf.reduce_max(pi_stddev, axis=-1) / tf.reduce_min(pi_stddev, axis=-1))

    # Log per-objective values.
    for i, name in enumerate(self._objective_names):
      stats["{}_dual_temperature".format(name)] = temperature[i]
      stats["{}_loss_temperature".format(name)] = loss_temperature[i]
      stats["{}_kl_q_rel".format(name)] = tf.reduce_mean(
          kl_nonparametric[:, i]) / self._epsilons[i]

      # Q measurements.
      stats["{}_q_min".format(name)] = tf.reduce_mean(tf.reduce_min(
          q_values, axis=0)[:, i])
      stats["{}_q_mean".format(name)] = tf.reduce_mean(tf.reduce_mean(
          q_values, axis=0)[:, i])
      stats["{}_q_max".format(name)] = tf.reduce_mean(tf.reduce_max(
          q_values, axis=0)[:, i])

    return loss, stats


def compute_weights_and_temperature_loss(
    q_values: tf.Tensor,
    epsilons: tf.Tensor,
    temperature: tf.Variable,
) -> Tuple[tf.Tensor, tf.Tensor]:
  """Computes normalized importance weights for the policy optimization.

  Args:
    q_values: Q-values associated with the actions sampled from the target
      policy; expected shape [N, B, K].
    epsilons: Desired per-objective constraints on the KL between the target
      and non-parametric policies; expected shape [K].
    temperature: Per-objective scalar used to temper the Q-values before
      computing normalized importance weights from them; expected shape [K].
      This is really the Lagrange dual variable in the constrained optimization
      problem, the solution of which is the non-parametric policy targeted by
      the policy loss.

  Returns:
    Normalized importance weights, used for policy optimization; shape [N,B,K].
    Temperature loss, used to adapt the temperature; shape [1, K].
  """

  # Temper the given Q-values using the current temperature.
  tempered_q_values = tf.stop_gradient(q_values) / temperature[None, None, :]

  # Compute the normalized importance weights used to compute expectations with
  # respect to the non-parametric policy.
  normalized_weights = tf.nn.softmax(tempered_q_values, axis=0)
  normalized_weights = tf.stop_gradient(normalized_weights)

  # Compute the temperature loss (dual of the E-step optimization problem).
  q_logsumexp = tf.reduce_logsumexp(tempered_q_values, axis=0)
  log_num_actions = tf.math.log(tf.cast(q_values.shape[0], tf.float32))
  loss_temperature = (
      epsilons + tf.reduce_mean(q_logsumexp, axis=0) - log_num_actions)
  loss_temperature = temperature * loss_temperature

  return normalized_weights, loss_temperature



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for R2D2 agent."""

from typing import Optional

from absl.testing import absltest
import acme
from acme import specs
from acme.agents.jax import r2d2
from acme.jax import networks as networks_lib
from acme.testing import fakes
from acme.utils import counting
import haiku as hk
import jax
import numpy as np


class RDQN(hk.RNNCore):
  """A simple recurrent network for testing."""

  def __init__(self, num_actions: int):
    super().__init__(name='my_network')
    self._torso = hk.Sequential([
        hk.Flatten(),
        hk.nets.MLP([50, 50]),
    ])
    self._core = hk.LSTM(20)
    self._head = networks_lib.PolicyValueHead(num_actions)

  def __call__(self, inputs, state):
    embeddings = self._torso(inputs)
    embeddings, new_state = self._core(embeddings, state)
    logits, _ = self._head(embeddings)
    return logits, new_state

  def initial_state(self, batch_size: int):
    return self._core.initial_state(batch_size)

  def unroll(self, inputs, state):
    embeddings = jax.vmap(self._torso)(inputs)  # [T D]
    embeddings, new_states = hk.static_unroll(self._core, embeddings, state)
    logits, _ = self._head(embeddings)
    return logits, new_states


def make_networks(spec, batch_size):
  """Creates networks used by the agent."""

  def forward_fn(inputs, state):
    model = RDQN(spec.actions.num_values)
    return model(inputs, state)

  def initial_state_fn(batch_size: Optional[int] = None):
    model = RDQN(spec.actions.num_values)
    return model.initial_state(batch_size)

  def unroll_fn(inputs, state):
    model = RDQN(spec.actions.num_values)
    return model.unroll(inputs, state)

  return r2d2.make_networks(
      forward_fn=forward_fn,
      initial_state_fn=initial_state_fn,
      unroll_fn=unroll_fn,
      env_spec=spec,
      batch_size=batch_size)


class R2D2Test(absltest.TestCase):

  def test_r2d2(self):
    # Create a fake environment to test with.
    environment = fakes.DiscreteEnvironment(
        num_actions=5,
        num_observations=10,
        obs_shape=(10, 5),
        obs_dtype=np.float32,
        episode_length=10)
    spec = specs.make_environment_spec(environment)

    config = r2d2.R2D2Config(
        batch_size=1,
        trace_length=5,
        sequence_period=1,
        samples_per_insert=0.,
        min_replay_size=1,
        burn_in_length=1)

    counter = counting.Counter()
    agent = r2d2.R2D2(
        spec=spec,
        networks=make_networks(spec, config.batch_size),
        config=config,
        seed=0,
        counter=counter,
    )

    # Try running the environment loop. We have no assertions here because all
    # we care about is that the agent runs without raising any errors.
    loop = acme.EnvironmentLoop(environment, agent, counter=counter)
    loop.run(num_episodes=20)


if __name__ == '__main__':
  absltest.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""IMPALA actor implementation."""

from typing import Optional

from acme import adders
from acme import core
from acme.agents.jax.impala import types
from acme.jax import variable_utils
import dm_env
import haiku as hk
import jax
import jax.numpy as jnp


class IMPALAActor(core.Actor):
  """A recurrent actor."""

  _state: hk.LSTMState
  _prev_state: hk.LSTMState
  _prev_logits: jnp.ndarray

  def __init__(
      self,
      forward_fn: types.PolicyValueFn,
      initial_state_init_fn: types.RecurrentStateInitFn,
      initial_state_fn: types.RecurrentStateFn,
      rng: hk.PRNGSequence,
      variable_client: Optional[variable_utils.VariableClient] = None,
      adder: Optional[adders.Adder] = None,
  ):

    # Store these for later use.
    self._adder = adder
    self._variable_client = variable_client
    self._forward = forward_fn
    self._reset_fn_or_none = getattr(forward_fn, 'reset', None)
    self._rng = rng

    # Make sure not to use a random policy after checkpoint restoration by
    # assigning variables before running the environment loop.
    if self._variable_client is not None:
      self._variable_client.update_and_wait()

    params = initial_state_init_fn(next(self._rng))
    self._initial_state = initial_state_fn(params)

  def select_action(self, observation: types.Observation) -> types.Action:

    if self._state is None:
      self._state = self._initial_state

    # Forward.
    (logits, _), new_state = self._forward(self._params, observation,
                                           self._state)

    self._prev_logits = logits
    self._prev_state = self._state
    self._state = new_state

    action = jax.random.categorical(next(self._rng), logits)

    return action

  def observe_first(self, timestep: dm_env.TimeStep):
    if self._adder:
      self._adder.add_first(timestep)

    # Set the state to None so that we re-initialize at the next policy call.
    self._state = None

    # Reset state of inference functions that employ stateful wrappers (eg. BIT)
    # at the start of the episode.
    if self._reset_fn_or_none is not None:
      self._reset_fn_or_none()

  def observe(
      self,
      action: types.Action,
      next_timestep: dm_env.TimeStep,
  ):
    if not self._adder:
      return

    extras = {'logits': self._prev_logits, 'core_state': self._prev_state}
    self._adder.add(action, next_timestep, extras)

  def update(self, wait: bool = False):
    if self._variable_client is not None:
      self._variable_client.update(wait)

  @property
  def _params(self) -> Optional[hk.Params]:
    if self._variable_client is None:
      # If self._variable_client is None then we assume self._forward  does not
      # use the parameters it is passed and just return None.
      return None
    return self._variable_client.params



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the AIL learner."""
import functools

from absl.testing import absltest
from acme import specs
from acme import types
from acme.agents.jax.ail import learning as ail_learning
from acme.agents.jax.ail import losses
from acme.agents.jax.ail import networks as ail_networks
from acme.jax import networks as networks_lib
from acme.jax import utils
import haiku as hk
import jax
from jax import test_util as jtu
import numpy as np
import optax


def _make_discriminator(spec):
  def discriminator(*args, **kwargs) -> networks_lib.Logits:
    return ail_networks.DiscriminatorModule(
        environment_spec=spec,
        use_action=False,
        use_next_obs=False,
        network_core=ail_networks.DiscriminatorMLP([]))(*args, **kwargs)

  discriminator_transformed = hk.without_apply_rng(
      hk.transform_with_state(discriminator))
  return ail_networks.make_discriminator(
      environment_spec=spec,
      discriminator_transformed=discriminator_transformed)


class AilLearnerTest(jtu.JaxTestCase):

  def test_step(self):
    simple_spec = specs.Array(shape=(), dtype=float)

    spec = specs.EnvironmentSpec(simple_spec, simple_spec, simple_spec,
                                 simple_spec)

    discriminator = _make_discriminator(spec)
    ail_network = ail_networks.AILNetworks(
        discriminator, imitation_reward_fn=lambda x: x, direct_rl_networks=None)

    loss = losses.gail_loss()

    optimizer = optax.adam(.01)

    step = jax.jit(functools.partial(
        ail_learning.ail_update_step,
        optimizer=optimizer,
        ail_network=ail_network,
        loss_fn=loss))

    zero_transition = types.Transition(
        np.array([0.]), np.array([0.]), 0., 0., np.array([0.]))
    zero_transition = utils.add_batch_dim(zero_transition)

    one_transition = types.Transition(
        np.array([1.]), np.array([0.]), 0., 0., np.array([0.]))
    one_transition = utils.add_batch_dim(one_transition)

    key = jax.random.PRNGKey(0)
    discriminator_params, discriminator_state = discriminator.init(key)

    state = ail_learning.DiscriminatorTrainingState(
        optimizer_state=optimizer.init(discriminator_params),
        discriminator_params=discriminator_params,
        discriminator_state=discriminator_state,
        policy_params=None,
        key=key,
        steps=0,
    )

    expected_loss = [1.062, 1.057, 1.052]

    for i in range(3):
      state, loss = step(state, (one_transition, zero_transition))
      self.assertAlmostEqual(loss['total_loss'], expected_loss[i], places=3)


if __name__ == '__main__':
  absltest.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""R2D2 Networks."""

import dataclasses

from typing import Any, Callable, Optional, Tuple

from acme import specs
from acme import types
from acme.agents.jax import actor_core as actor_core_lib
from acme.agents.jax.r2d2 import config as r2d2_config
from acme.jax import networks as networks_lib
from acme.jax import utils
import haiku as hk
import rlax


Epsilon = float
EpsilonRecurrentPolicy = Callable[[
    networks_lib.Params, networks_lib.PRNGKey, networks_lib
    .Observation, actor_core_lib.RecurrentState, Epsilon
], Tuple[networks_lib.Action, actor_core_lib.RecurrentState]]


@dataclasses.dataclass
class R2D2Networks:
  """Network and pure functions for the R2D2 agent.."""
  forward: networks_lib.FeedForwardNetwork
  unroll: networks_lib.FeedForwardNetwork
  initial_state: networks_lib.FeedForwardNetwork


def make_networks(
    env_spec: specs.EnvironmentSpec,
    forward_fn: Any,
    initial_state_fn: Any,
    unroll_fn: Any,
    batch_size) -> R2D2Networks:
  """Builds functional r2d2 network from recurrent model definitions."""

  # Make networks purely functional.
  forward_hk = hk.transform(forward_fn)
  initial_state_hk = hk.transform(initial_state_fn)
  unroll_hk = hk.transform(unroll_fn)

  # Define networks init functions.
  def initial_state_init_fn(rng, batch_size):
    return initial_state_hk.init(rng, batch_size)
  dummy_obs_batch = utils.tile_nested(
      utils.zeros_like(env_spec.observations), batch_size)
  dummy_obs_sequence = utils.add_batch_dim(dummy_obs_batch)
  def unroll_init_fn(rng, initial_state):
    return unroll_hk.init(rng, dummy_obs_sequence, initial_state)

  # Make FeedForwardNetworks.
  forward = networks_lib.FeedForwardNetwork(
      init=forward_hk.init, apply=forward_hk.apply)
  unroll = networks_lib.FeedForwardNetwork(
      init=unroll_init_fn, apply=unroll_hk.apply)
  initial_state = networks_lib.FeedForwardNetwork(
      init=initial_state_init_fn, apply=initial_state_hk.apply)
  return R2D2Networks(
      forward=forward, unroll=unroll, initial_state=initial_state)


def make_atari_networks(batch_size, env_spec):
  """Builds default R2D2 networks for Atari games."""

  def forward_fn(x, s):
    model = networks_lib.R2D2AtariNetwork(env_spec.actions.num_values)
    return model(x, s)

  def initial_state_fn(batch_size: Optional[int] = None):
    model = networks_lib.R2D2AtariNetwork(env_spec.actions.num_values)
    return model.initial_state(batch_size)

  def unroll_fn(inputs, state):
    model = networks_lib.R2D2AtariNetwork(env_spec.actions.num_values)
    return model.unroll(inputs, state)

  return make_networks(env_spec=env_spec, forward_fn=forward_fn,
                       initial_state_fn=initial_state_fn, unroll_fn=unroll_fn,
                       batch_size=batch_size)


def make_behavior_policy(
    networks: R2D2Networks,
    config: r2d2_config.R2D2Config,
    evaluation: bool = False) -> EpsilonRecurrentPolicy:
  """Selects action according to the policy."""

  def behavior_policy(params: networks_lib.Params, key: networks_lib.PRNGKey,
                      observation: types.NestedArray,
                      core_state: types.NestedArray,
                      epsilon):
    q_values, core_state = networks.forward.apply(
        params, key, observation, core_state)
    epsilon = config.evaluation_epsilon if evaluation else epsilon
    return rlax.epsilon_greedy(epsilon).sample(key, q_values), core_state

  return behavior_policy



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for nested data structures involving NumPy and TensorFlow 2.x."""

import functools
from typing import List, Optional, Union

from acme import types
from acme.utils import tree_utils

import sonnet as snt
import tensorflow as tf
import tree


def add_batch_dim(nest: types.NestedTensor) -> types.NestedTensor:
  """Adds a batch dimension to each leaf of a nested structure of Tensors."""
  return tree.map_structure(lambda x: tf.expand_dims(x, axis=0), nest)


def squeeze_batch_dim(nest: types.NestedTensor) -> types.NestedTensor:
  """Squeezes out a batch dimension from each leaf of a nested structure."""
  return tree.map_structure(lambda x: tf.squeeze(x, axis=0), nest)


def batch_concat(inputs: types.NestedTensor) -> tf.Tensor:
  """Concatenate a collection of Tensors while preserving the batch dimension.

  This takes a potentially nested collection of tensors, flattens everything
  but the batch (first) dimension, and concatenates along the resulting data
  (second) dimension.

  Args:
    inputs: a tensor or nested collection of tensors.

  Returns:
    A concatenated tensor which maintains the batch dimension but concatenates
    all other data along the flattened second dimension.
  """
  flat_leaves = tree.map_structure(snt.Flatten(), inputs)
  return tf.concat(tree.flatten(flat_leaves), axis=-1)


def batch_to_sequence(data: types.NestedTensor) -> types.NestedTensor:
  """Converts data between sequence-major and batch-major format."""
  return tree.map_structure(
      lambda t: tf.transpose(t, [1, 0] + list(range(2, t.shape.rank))), data)


def tile_tensor(tensor: tf.Tensor, multiple: int) -> tf.Tensor:
  """Tiles `multiple` copies of `tensor` along a new leading axis."""
  rank = len(tensor.shape)
  multiples = tf.constant([multiple] + [1] * rank, dtype=tf.int32)
  expanded_tensor = tf.expand_dims(tensor, axis=0)
  return tf.tile(expanded_tensor, multiples)


def tile_nested(inputs: types.NestedTensor,
                multiple: int) -> types.NestedTensor:
  """Tiles tensors in a nested structure along a new leading axis."""
  tile = functools.partial(tile_tensor, multiple=multiple)
  return tree.map_structure(tile, inputs)


def create_variables(
    network: snt.Module,
    input_spec: List[Union[types.NestedSpec, tf.TensorSpec]],
) -> Optional[tf.TensorSpec]:
  """Builds the network with dummy inputs to create the necessary variables.

  Args:
    network: Sonnet Module whose variables are to be created.
    input_spec: list of input specs to the network. The length of this list
      should match the number of arguments expected by `network`.

  Returns:
    output_spec: only returns an output spec if the output is a tf.Tensor, else
        it doesn't return anything (None); e.g. if the output is a
        tfp.distributions.Distribution.
  """
  # Create a dummy observation with no batch dimension.
  dummy_input = zeros_like(input_spec)

  # If we have an RNNCore the hidden state will be an additional input.
  if isinstance(network, snt.RNNCore):
    initial_state = squeeze_batch_dim(network.initial_state(1))
    dummy_input += [initial_state]

  # Forward pass of the network which will create variables as a side effect.
  dummy_output = network(*add_batch_dim(dummy_input))

  # Evaluate the input signature by converting the dummy input into a
  # TensorSpec. We then save the signature as a property of the network. This is
  # done so that we can later use it when creating snapshots. We do this here
  # because the snapshot code may not have access to the precise form of the
  # inputs.
  input_signature = tree.map_structure(
      lambda t: tf.TensorSpec((None,) + t.shape, t.dtype), dummy_input)
  network._input_signature = input_signature  # pylint: disable=protected-access

  def spec(output):
    # If the output is not a Tensor, return None as spec is ill-defined.
    if not isinstance(output, tf.Tensor):
      return None
    # If this is not a scalar Tensor, make sure to squeeze out the batch dim.
    if tf.rank(output) > 0:
      output = squeeze_batch_dim(output)
    return tf.TensorSpec(output.shape, output.dtype)

  return tree.map_structure(spec, dummy_output)


class TransformationWrapper(snt.Module):
  """Helper class for to_sonnet_module.

  This wraps arbitrary Tensor-valued callables as a Sonnet module.
  A use case for this is in agent code that could take either a trainable
  sonnet module or a hard-coded function as its policy. By wrapping a hard-coded
  policy with this class, the agent can then treat it as if it were a Sonnet
  module. This removes the need for "if is_hard_coded:..." branches, which you'd
  otherwise need if e.g. calling get_variables() on the policy.
  """

  def __init__(self,
               transformation: types.TensorValuedCallable,
               name: Optional[str] = None):
    super().__init__(name=name)
    self._transformation = transformation

  def __call__(self, *args, **kwargs):
    return self._transformation(*args, **kwargs)


def to_sonnet_module(
    transformation: types.TensorValuedCallable
    ) -> snt.Module:
  """Convert a tensor transformation to a Sonnet Module.

  Args:
    transformation: A Callable that takes one or more (nested) Tensors, and
      returns one or more (nested) Tensors.

  Returns:
    A Sonnet Module that wraps the transformation.
  """

  if isinstance(transformation, snt.Module):
    return transformation

  module = TransformationWrapper(transformation)

  # Wrap the module to allow it to return an empty variable tuple.
  return snt.allow_empty_variables(module)


def to_numpy(nest: types.NestedTensor) -> types.NestedArray:
  """Converts a nest of Tensors to a nest of numpy arrays."""
  return tree.map_structure(lambda x: x.numpy(), nest)


def to_numpy_squeeze(nest: types.NestedTensor, axis=0) -> types.NestedArray:
  """Converts a nest of Tensors to a nest of numpy arrays and squeeze axis."""
  return tree.map_structure(lambda x: x.numpy().squeeze(axis=axis), nest)


def zeros_like(nest: types.Nest) -> types.NestedTensor:
  """Given a nest of array-like objects, returns similarly nested tf.zeros."""
  return tree.map_structure(lambda x: tf.zeros(x.shape, x.dtype), nest)


# TODO(b/160311329): Migrate call-sites and remove.
stack_sequence_fields = tree_utils.stack_sequence_fields



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility classes for input normalization."""

import dataclasses
import functools
from typing import Any, Callable, Iterator, List, Optional, Tuple

import acme
from acme import adders
from acme import core
from acme import specs
from acme import types
from acme.agents.jax import builders
from acme.jax import networks as networks_lib
from acme.jax import running_statistics
from acme.jax import variable_utils
from acme.utils import counting
import dm_env
import jax
import reverb

_NORMALIZATION_VARIABLES = 'normalization_variables'


# Wrapping the network instead might look more straightforward, but then
# different implementations would be needed for feed-forward and
# recurrent networks.
class NormalizationActorWrapper(core.Actor):
  """An actor wrapper that normalizes observations before applying policy."""

  def __init__(self,
               wrapped_actor: core.Actor,
               variable_source: core.VariableSource,
               max_abs_observation: Optional[float],
               update_period: int = 1,
               backend: Optional[str] = None):
    self._wrapped_actor = wrapped_actor
    self._variable_client = variable_utils.VariableClient(
        variable_source,
        key=_NORMALIZATION_VARIABLES,
        update_period=update_period,
        device=backend)
    self._apply_normalization = jax.jit(
        functools.partial(
            running_statistics.normalize, max_abs_value=max_abs_observation),
        backend=backend)

  def select_action(self, observation: types.NestedArray) -> types.NestedArray:
    self._variable_client.update()
    observation_stats = self._variable_client.params
    observation = self._apply_normalization(observation, observation_stats)
    return self._wrapped_actor.select_action(observation)

  def observe_first(self, timestep: dm_env.TimeStep):
    return self._wrapped_actor.observe_first(timestep)

  def observe(
      self,
      action: types.NestedArray,
      next_timestep: dm_env.TimeStep,
  ):
    return self._wrapped_actor.observe(action, next_timestep)

  def update(self, wait: bool = False):
    return self._wrapped_actor.update(wait)


@dataclasses.dataclass
class NormalizationLearnerWrapperState:
  wrapped_learner_state: Any
  observation_running_statistics: running_statistics.RunningStatisticsState


class NormalizationLearnerWrapper(core.Learner, core.Saveable):
  """A learner wrapper that normalizes observations using running statistics."""

  def __init__(self, learner_factory: Callable[[Iterator[reverb.ReplaySample]],
                                               acme.Learner],
               iterator: Iterator[reverb.ReplaySample],
               environment_spec: specs.EnvironmentSpec, is_sequence_based: bool,
               batch_dims: Optional[Tuple[int, ...]],
               max_abs_observation: Optional[float]):

    def normalize_sample(
        observation_statistics: running_statistics.RunningStatisticsState,
        sample: reverb.ReplaySample
    ) -> Tuple[running_statistics.RunningStatisticsState, reverb.ReplaySample]:
      observation = sample.data.observation
      observation_statistics = running_statistics.update(
          observation_statistics, observation)
      observation = running_statistics.normalize(
          observation,
          observation_statistics,
          max_abs_value=max_abs_observation)
      if is_sequence_based:
        assert not hasattr(sample.data, 'next_observation')
        sample = reverb.ReplaySample(
            sample.info, sample.data._replace(observation=observation))
      else:
        next_observation = running_statistics.normalize(
            sample.data.next_observation,
            observation_statistics,
            max_abs_value=max_abs_observation)
        sample = reverb.ReplaySample(
            sample.info,
            sample.data._replace(
                observation=observation, next_observation=next_observation))

      return observation_statistics, sample

    self._observation_running_statistics = running_statistics.init_state(
        environment_spec.observations)
    self._normalize_sample = jax.jit(normalize_sample)

    normalizing_iterator = (
        self._normalize_sample_and_update(sample) for sample in iterator)
    self._wrapped_learner = learner_factory(normalizing_iterator)

  def _normalize_sample_and_update(
      self, sample: reverb.ReplaySample) -> reverb.ReplaySample:
    self._observation_running_statistics, sample = self._normalize_sample(
        self._observation_running_statistics, sample)
    return sample

  def step(self):
    self._wrapped_learner.step()

  def get_variables(self, names: List[str]) -> List[types.NestedArray]:
    stats = self._observation_running_statistics
    # Make sure to only pass mean and std to minimize trafic.
    mean_std = running_statistics.NestedMeanStd(mean=stats.mean, std=stats.std)
    normalization_variables = {_NORMALIZATION_VARIABLES: mean_std}

    learner_names = [
        name for name in names if name not in normalization_variables
    ]
    learner_variables = dict(
        zip(learner_names, self._wrapped_learner.get_variables(
            learner_names))) if learner_names else {}

    return [
        normalization_variables.get(name, learner_variables.get(name, None))
        for name in names
    ]

  def save(self) -> NormalizationLearnerWrapperState:
    return NormalizationLearnerWrapperState(
        wrapped_learner_state=self._wrapped_learner.save(),
        observation_running_statistics=self._observation_running_statistics)

  def restore(self, state: NormalizationLearnerWrapperState):
    self._wrapped_learner.restore(state.wrapped_learner_state)
    self._observation_running_statistics = state.observation_running_statistics


@dataclasses.dataclass
class NormalizationBuilder(builders.ActorLearnerBuilder):
  """Builder wrapper that normalizes observations using running mean/std."""
  builder: builders.ActorLearnerBuilder
  environment_spec: specs.EnvironmentSpec
  is_sequence_based: bool
  batch_dims: Optional[Tuple[int, ...]]
  max_abs_observation: Optional[float] = 10.0
  statistics_update_period: int = 100

  def make_replay_tables(
      self, environment_spec: specs.EnvironmentSpec) -> List[reverb.Table]:
    return self.builder.make_replay_tables(environment_spec)

  def make_dataset_iterator(
      self, replay_client: reverb.Client) -> Iterator[reverb.ReplaySample]:
    return self.builder.make_dataset_iterator(replay_client)

  def make_adder(self,
                 replay_client: reverb.Client) -> Optional[adders.Adder]:
    return self.builder.make_adder(replay_client)

  def make_learner(
      self,
      random_key: networks_lib.PRNGKey,
      networks,
      dataset: Iterator[reverb.ReplaySample],
      replay_client: Optional[reverb.Client] = None,
      counter: Optional[counting.Counter] = None,
  ) -> core.Learner:

    learner_factory = functools.partial(
        self.builder.make_learner,
        random_key,
        networks,
        replay_client=replay_client,
        counter=counter)

    return NormalizationLearnerWrapper(
        learner_factory=learner_factory,
        iterator=dataset,
        environment_spec=self.environment_spec,
        is_sequence_based=self.is_sequence_based,
        batch_dims=self.batch_dims,
        max_abs_observation=self.max_abs_observation)

  def make_actor(
      self,
      random_key: networks_lib.PRNGKey,
      policy_network,
      adder: Optional[adders.Adder] = None,
      variable_source: Optional[core.VariableSource] = None,
  ) -> core.Actor:
    actor = self.builder.make_actor(random_key, policy_network, adder,
                                    variable_source)
    return NormalizationActorWrapper(
        actor,
        variable_source,
        max_abs_observation=self.max_abs_observation,
        update_period=self.statistics_update_period,
        backend='cpu')



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Adversarial Imitation Learning (AIL) Builder."""

import functools
import itertools
from typing import Callable, Generic, Iterator, List, Optional, Tuple

from acme import adders
from acme import core
from acme import specs
from acme import types
from acme.adders import reverb as adders_reverb
from acme.agents.jax import builders
from acme.agents.jax.ail import config as ail_config
from acme.agents.jax.ail import learning
from acme.agents.jax.ail import losses
from acme.agents.jax.ail import networks as ail_networks
from acme.datasets import reverb as datasets
from acme.jax import networks as networks_lib
from acme.jax.imitation_learning_types import DirectPolicyNetwork
from acme.jax.types import PRNGKey
from acme.utils import counting
from acme.utils import loggers
from acme.utils import reverb_utils
import jax
import numpy as np
import optax
import reverb
from reverb import rate_limiters
import tree


def _split_transitions(
    transitions: types.Transition,
    index: int) -> Tuple[types.Transition, types.Transition]:
  """Splits the given transition on the first axis at the given index.

  Args:
    transitions: Transitions to split.
    index: Spliting index.

  Returns:
    A pair of transitions, the first containing elements before the index
    (exclusive) and the second after the index (inclusive)
  """
  return (tree.map_structure(lambda x: x[:index], transitions),
          tree.map_structure(lambda x: x[index:], transitions))


def _rebatch(iterator: Iterator[types.Transition],
             batch_size: int) -> Iterator[types.Transition]:
  """Rebatch the itererator with the given batch size.

  Args:
    iterator: Iterator to rebatch.
    batch_size: New batch size.

  Yields:
    Transitions with the new batch size.
  """
  data = next(iterator)
  while True:
    while len(data.reward) < batch_size:
      # Ensure we can get enough demonstrations.
      next_data = next(iterator)
      data = tree.map_structure(lambda *args: np.concatenate(list(args)), data,
                                next_data)
    output, data = _split_transitions(data, batch_size)
    yield output


def _mix_arrays(
    replay: np.ndarray,
    demo: np.ndarray,
    index: int,
    seed: int) -> np.ndarray:
  """Mixes `replay` and `demo`.

  Args:
    replay: Replay data to mix. Only index element will be selected.
    demo: Demonstration data to mix.
    index: Amount of replay data we should include.
    seed: RNG seed.

  Returns:
    An array with replay elements up to 'index' and all the demos.
  """
  # We're throwing away some replay data here. We have to if we want to make
  # sure the output info field is correct.
  output = np.concatenate((replay[:index], demo))
  return np.random.default_rng(seed=seed).permutation(output)


def _generate_samples_with_demonstrations(
    demonstration_iterator: Iterator[types.Transition],
    replay_iterator: Iterator[reverb.ReplaySample],
    policy_to_expert_data_ratio: int,
    batch_size) -> Iterator[reverb.ReplaySample]:
  """Generator which creates the sample having demonstrations in them.

  It takes the demonstrations and replay iterators and generates batches with
  same size as the replay iterator, such that each batches have the ratio of
  policy and expert data specified in policy_to_expert_data_ratio on average.
  There is no constraints on how the demonstrations and replay samples should be
  batched.

  Args:
    demonstration_iterator: Iterator of demonstrations.
    replay_iterator: Replay buffer sample iterator.
    policy_to_expert_data_ratio: Amount of policy transitions for 1 expert
      transitions.
    batch_size: Output batch size, which should match the replay batch size.

  Yields:
    Samples having a mix of demonstrations and policy data. The info will match
    the current replay sample info and the batch size will be the same as the
    replay_iterator data batch size.
  """
  count = 0
  if batch_size % (policy_to_expert_data_ratio + 1) != 0:
    raise ValueError(
        'policy_to_expert_data_ratio + 1 must divide the batch size but '
        f'{batch_size} % {policy_to_expert_data_ratio+1} !=0')
  demo_insertion_size = batch_size // (policy_to_expert_data_ratio + 1)
  policy_insertion_size = batch_size - demo_insertion_size

  demonstration_iterator = _rebatch(demonstration_iterator, demo_insertion_size)
  for sample, demos in zip(replay_iterator, demonstration_iterator):
    output_transitions = tree.map_structure(
        functools.partial(_mix_arrays,
                          index=policy_insertion_size,
                          seed=count),
        sample.data, demos)
    count += 1
    yield reverb.ReplaySample(info=sample.info, data=output_transitions)


class AILBuilder(builders.GenericActorLearnerBuilder[ail_networks.AILNetworks,
                                                     DirectPolicyNetwork,
                                                     learning.AILSample],
                 Generic[ail_networks.DirectRLNetworks, DirectPolicyNetwork]):
  """AIL Builder."""

  def __init__(self,
               rl_agent: builders.GenericActorLearnerBuilder[
                   ail_networks.DirectRLNetworks, DirectPolicyNetwork,
                   reverb.ReplaySample],
               config: ail_config.AILConfig,
               discriminator_loss: losses.Loss,
               make_demonstrations: Callable[[int], Iterator[types.Transition]],
               logger_fn: Callable[[], loggers.Logger] = lambda: None):
    """Implements a builder for AIL using rl_agent as forward RL algorithm.

    Args:
      rl_agent: The standard RL agent used by AIL to optimize the generator.
      config: a AIL config
      discriminator_loss: The loss function for the discriminator to minimize.
      make_demonstrations: A function that returns an iterator with
        demonstrations to be imitated.
      logger_fn: a logger factory for the learner
    """
    self._rl_agent = rl_agent
    self._config = config
    self._discriminator_loss = discriminator_loss
    self._make_demonstrations = make_demonstrations
    self._logger_fn = logger_fn

  def make_learner(self,
                   random_key: networks_lib.PRNGKey,
                   networks: ail_networks.AILNetworks,
                   dataset: Iterator[learning.AILSample],
                   replay_client: Optional[reverb.Client] = None,
                   counter: Optional[counting.Counter] = None) -> core.Learner:
    counter = counter or counting.Counter()
    direct_rl_counter = counting.Counter(counter, 'direct_rl')
    batch_size_per_learner_step = ail_config.get_per_learner_step_batch_size(
        self._config)

    direct_rl_learner_key, discriminator_key = jax.random.split(random_key)

    direct_rl_learner = functools.partial(
        self._rl_agent.make_learner,
        direct_rl_learner_key,
        networks.direct_rl_networks,
        replay_client=replay_client,
        counter=direct_rl_counter)

    discriminator_optimizer = (
        self._config.discriminator_optimizer or optax.adam(1e-5))

    return learning.AILLearner(
        counter,
        direct_rl_learner_factory=direct_rl_learner,
        loss_fn=self._discriminator_loss,
        iterator=dataset,
        discriminator_optimizer=discriminator_optimizer,
        ail_network=networks,
        discriminator_key=discriminator_key,
        is_sequence_based=self._config.is_sequence_based,
        num_sgd_steps_per_step=batch_size_per_learner_step //
        self._config.discriminator_batch_size,
        policy_variable_name=self._config.policy_variable_name,
        logger=self._logger_fn())

  def make_replay_tables(
      self, environment_spec: specs.EnvironmentSpec) -> List[reverb.Table]:
    replay_tables = self._rl_agent.make_replay_tables(environment_spec)
    if self._config.share_iterator:
      return replay_tables
    replay_tables.append(
        reverb.Table(
            name=self._config.replay_table_name,
            sampler=reverb.selectors.Uniform(),
            remover=reverb.selectors.Fifo(),
            max_size=self._config.max_replay_size,
            rate_limiter=rate_limiters.MinSize(self._config.min_replay_size),
            signature=adders_reverb.NStepTransitionAdder.signature(
                environment_spec)))
    return replay_tables

  # This function does not expose all the iterators used by the learner when
  # share_iterator is False, making further wrapping impossible.
  # TODO(eorsini): Expose all iterators.
  # Currently GAIL uses 3 iterators, instead we can make it use a single
  # iterator and return this one here. The way to achieve this would be:
  # * Create the 3 iterators here.
  # * zip them and return them here.
  # * upzip them in the learner (this step will not be necessary once we move to
  # stateless learners)
  # This should work fine as the 3 iterators are always iterated in parallel
  # (i.e. at every step we call next once on each of them).
  def make_dataset_iterator(
      self, replay_client: reverb.Client) -> Iterator[learning.AILSample]:
    batch_size_per_learner_step = ail_config.get_per_learner_step_batch_size(
        self._config)

    iterator_demonstration = self._make_demonstrations(
        batch_size_per_learner_step)

    direct_iterator = self._rl_agent.make_dataset_iterator(replay_client)

    if self._config.share_iterator:
      # In order to reuse the iterator return values and not lose a 2x factor on
      # sample efficiency, we need to use itertools.tee().
      discriminator_iterator, direct_iterator = itertools.tee(direct_iterator)
    else:
      discriminator_iterator = datasets.make_reverb_dataset(
          table=self._config.replay_table_name,
          server_address=replay_client.server_address,
          batch_size=ail_config.get_per_learner_step_batch_size(self._config),
          prefetch_size=self._config.prefetch_size).as_numpy_iterator()

    if self._config.policy_to_expert_data_ratio is not None:
      iterator_demonstration, iterator_demonstration2 = itertools.tee(
          iterator_demonstration)
      direct_iterator = _generate_samples_with_demonstrations(
          iterator_demonstration2, direct_iterator,
          self._config.policy_to_expert_data_ratio,
          self._config.direct_rl_batch_size)

    is_sequence_based = self._config.is_sequence_based

    # Don't flatten the discriminator batch if the iterator is not shared.
    process_discriminator_sample = functools.partial(
        reverb_utils.replay_sample_to_sars_transition,
        is_sequence=is_sequence_based and self._config.share_iterator,
        flatten_batch=is_sequence_based and self._config.share_iterator,
        strip_last_transition=is_sequence_based and self._config.share_iterator)

    discriminator_iterator = (
        process_discriminator_sample(sample)
        for sample in discriminator_iterator)

    return (learning.AILSample(*sample) for sample in zip(
        discriminator_iterator, direct_iterator, iterator_demonstration))

  def make_adder(self,
                 replay_client: reverb.Client) -> Optional[adders.Adder]:
    direct_rl_adder = self._rl_agent.make_adder(replay_client)
    if self._config.share_iterator:
      return direct_rl_adder
    ail_adder = adders_reverb.NStepTransitionAdder(
        priority_fns={self._config.replay_table_name: None},
        client=replay_client,
        n_step=1,
        discount=self._config.discount)

    # Some direct rl algorithms (such as PPO), might be passing extra data
    # which we won't be able to process here properly, so we need to ignore them
    return adders.ForkingAdder(
        [adders.IgnoreExtrasAdder(ail_adder), direct_rl_adder])

  def make_actor(
      self,
      random_key: PRNGKey,
      policy_network: DirectPolicyNetwork,
      adder: Optional[adders.Adder] = None,
      variable_source: Optional[core.VariableSource] = None,
  ) -> core.Actor:
    return self._rl_agent.make_actor(random_key, policy_network, adder,
                                     variable_source)



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Environment wrappers which record videos.

The code used to generate animations in this wrapper is based on that used in
the `dm_control/tutorial.ipynb` file.
"""

import os.path
from typing import Callable, Optional, Sequence, Tuple, Union
from acme.utils import paths
from acme.wrappers import base
import dm_env

import matplotlib
matplotlib.use('Agg')  # Switch to headless 'Agg' to inhibit figure rendering.
import matplotlib.animation as anim  # pylint: disable=g-import-not-at-top
import matplotlib.pyplot as plt
import numpy as np

# Internal imports.
# Make sure you have FFMpeg configured.

def make_animation(frames: Sequence[np.ndarray], frame_rate: float,
                   figsize: Union[float, Tuple[int, int]]) -> anim.Animation:
  """Generates an animation from a stack of frames."""

  # Set animation characteristics.
  if figsize is None:
    height, width, _ = frames[0].shape
  elif isinstance(figsize, tuple):
    height, width = figsize
  else:
    diagonal = figsize
    height, width, _ = frames[0].shape
    scale_factor = diagonal / np.sqrt(height**2 + width**2)
    width *= scale_factor
    height *= scale_factor

  dpi = 70
  interval = int(round(1e3 / frame_rate))  # Time (in ms) between frames.

  # Create and configure the figure.
  fig, ax = plt.subplots(1, 1, figsize=(width / dpi, height / dpi), dpi=dpi)
  ax.set_axis_off()
  ax.set_aspect('equal')
  ax.set_position([0, 0, 1, 1])

  # Initialize the first frame.
  im = ax.imshow(frames[0])

  # Create the function that will modify the frame, creating an animation.
  def update(frame):
    im.set_data(frame)
    return [im]

  return anim.FuncAnimation(
      fig=fig,
      func=update,
      frames=frames,
      interval=interval,
      blit=True,
      repeat=False)


class VideoWrapper(base.EnvironmentWrapper):
  """Wrapper which creates and records videos from generated observations.

  This will limit itself to recording once every `record_every` episodes and
  videos will be recorded to the directory `path` + '/<unique id>/videos' where
  `path` defaults to '~/acme'. Users can specify the size of the screen by
  passing either a tuple giving height and width or a float giving the size
  of the diagonal.
  """

  def __init__(self,
               environment: dm_env.Environment,
               *,
               path: str = '~/acme',
               filename: str = '',
               process_path: Callable[[str, str], str] = paths.process_path,
               record_every: int = 100,
               frame_rate: int = 30,
               figsize: Optional[Union[float, Tuple[int, int]]] = None):
    super(VideoWrapper, self).__init__(environment)
    self._path = process_path(path, 'videos')
    self._filename = filename
    self._record_every = record_every
    self._frame_rate = frame_rate
    self._frames = []
    self._counter = 0
    self._figsize = figsize

  def _render_frame(self, observation):
    """Renders a frame from the given environment observation."""
    return observation

  def _write_frames(self):
    """Writes frames to video."""
    if self._counter % self._record_every == 0:
      path = os.path.join(self._path,
                          f'{self._filename}_{self._counter:04d}.html')
      video = make_animation(self._frames, self._frame_rate,
                             self._figsize).to_html5_video()

      with open(path, 'w') as f:
        f.write(video)

    # Clear the frame buffer whether a video was generated or not.
    self._frames = []

  def _append_frame(self, observation):
    """Appends a frame to the sequence of frames."""
    if self._counter % self._record_every == 0:
      self._frames.append(self._render_frame(observation))

  def step(self, action) -> dm_env.TimeStep:
    timestep = self.environment.step(action)
    self._append_frame(timestep.observation)
    return timestep

  def reset(self) -> dm_env.TimeStep:
    # If the frame buffer is nonempty, flush it and record video
    if self._frames:
      self._write_frames()
    self._counter += 1
    timestep = self.environment.reset()
    self._append_frame(timestep.observation)
    return timestep

  def make_html_animation(self):
    if self._frames:
      return make_animation(self._frames, self._frame_rate,
                            self._figsize).to_html5_video()
    else:
      raise ValueError('make_html_animation should be called after running a '
                       'trajectory and before calling reset().')


class MujocoVideoWrapper(VideoWrapper):
  """VideoWrapper which generates videos from a mujoco physics object.

  This passes its keyword arguments into the parent `VideoWrapper` class (refer
  here for any default arguments).
  """

  # Note that since we can be given a wrapped mujoco environment we can't give
  # the type as dm_control.Environment.

  def __init__(self,
               environment: dm_env.Environment,
               *,
               frame_rate: Optional[int] = None,
               camera_id: Optional[int] = 0,
               height: int = 240,
               width: int = 320,
               playback_speed: float = 1.,
               **kwargs):

    # Check that we have a mujoco environment (or a wrapper thereof).
    if not hasattr(environment, 'physics'):
      raise ValueError('MujocoVideoWrapper expects an environment which '
                       'exposes a physics attribute corresponding to a MuJoCo '
                       'physics engine')

    # Compute frame rate if not set.
    if frame_rate is None:
      try:
        control_timestep = getattr(environment, 'control_timestep')()
      except AttributeError as e:
        raise AttributeError('MujocoVideoWrapper expects an environment which '
                             'exposes a control_timestep method, like '
                             'dm_control environments, or frame_rate '
                             'to be specified.') from e
      frame_rate = int(round(playback_speed / control_timestep))

    super().__init__(environment, frame_rate=frame_rate, **kwargs)
    self._camera_id = camera_id
    self._height = height
    self._width = width

  def _render_frame(self, unused_observation):
    del unused_observation

    # We've checked above that this attribute should exist. Pytype won't like
    # it if we just try and do self.environment.physics, so we use the slightly
    # grosser version below.
    physics = getattr(self.environment, 'physics')

    if self._camera_id is not None:
      frame = physics.render(
          camera_id=self._camera_id, height=self._height, width=self._width)
    else:
      # If camera_id is None, we create a minimal canvas that will accommodate
      # physics.model.ncam frames, and render all of them on a grid.
      num_cameras = physics.model.ncam
      num_columns = int(np.ceil(np.sqrt(num_cameras)))
      num_rows = int(np.ceil(float(num_cameras)/num_columns))
      height = self._height
      width = self._width

      # Make a black canvas.
      frame = np.zeros((num_rows*height, num_columns*width, 3), dtype=np.uint8)

      for col in range(num_columns):
        for row in range(num_rows):

          camera_id = row*num_columns + col

          if camera_id >= num_cameras:
            break

          subframe = physics.render(
              camera_id=camera_id, height=height, width=width)

          # Place the frame in the appropriate rectangle on the pixel canvas.
          frame[row*height:(row+1)*height, col*width:(col+1)*width] = subframe

    return frame



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DQfD Agent implementation."""

import copy
import functools
import operator
from typing import Optional

from acme import datasets
from acme import specs
from acme import types as acme_types
from acme.adders import reverb as adders
from acme.agents import agent
from acme.agents.tf import actors
from acme.agents.tf import dqn
from acme.tf import utils as tf2_utils
import reverb
import sonnet as snt
import tensorflow as tf
import tree
import trfl


class DQfD(agent.Agent):
  """DQfD agent.

  This implements a single-process DQN agent that mixes demonstrations with
  actor experience.
  """

  def __init__(
      self,
      environment_spec: specs.EnvironmentSpec,
      network: snt.Module,
      demonstration_dataset: tf.data.Dataset,
      demonstration_ratio: float,
      batch_size: int = 256,
      prefetch_size: int = 4,
      target_update_period: int = 100,
      samples_per_insert: float = 32.0,
      min_replay_size: int = 1000,
      max_replay_size: int = 1000000,
      importance_sampling_exponent: float = 0.2,
      n_step: int = 5,
      epsilon: Optional[tf.Tensor] = None,
      learning_rate: float = 1e-3,
      discount: float = 0.99,
  ):
    """Initialize the agent.

    Args:
      environment_spec: description of the actions, observations, etc.
      network: the online Q network (the one being optimized)
      demonstration_dataset: tf.data.Dataset producing (timestep, action)
        tuples containing full episodes.
      demonstration_ratio: Ratio of transitions coming from demonstrations.
      batch_size: batch size for updates.
      prefetch_size: size to prefetch from replay.
      target_update_period: number of learner steps to perform before updating
        the target networks.
      samples_per_insert: number of samples to take from replay for every insert
        that is made.
      min_replay_size: minimum replay size before updating. This and all
        following arguments are related to dataset construction and will be
        ignored if a dataset argument is passed.
      max_replay_size: maximum replay size.
      importance_sampling_exponent: power to which importance weights are raised
        before normalizing.
      n_step: number of steps to squash into a single transition.
      epsilon: probability of taking a random action; ignored if a policy
        network is given.
      learning_rate: learning rate for the q-network update.
      discount: discount to use for TD updates.
    """

    # Create a replay server to add data to. This uses no limiter behavior in
    # order to allow the Agent interface to handle it.
    replay_table = reverb.Table(
        name=adders.DEFAULT_PRIORITY_TABLE,
        sampler=reverb.selectors.Uniform(),
        remover=reverb.selectors.Fifo(),
        max_size=max_replay_size,
        rate_limiter=reverb.rate_limiters.MinSize(1),
        signature=adders.NStepTransitionAdder.signature(environment_spec))
    self._server = reverb.Server([replay_table], port=None)

    # The adder is used to insert observations into replay.
    address = f'localhost:{self._server.port}'
    adder = adders.NStepTransitionAdder(
        client=reverb.Client(address),
        n_step=n_step,
        discount=discount)

    # The dataset provides an interface to sample from replay.
    replay_client = reverb.TFClient(address)
    dataset = datasets.make_reverb_dataset(server_address=address)

    # Combine with demonstration dataset.
    transition = functools.partial(_n_step_transition_from_episode,
                                   n_step=n_step,
                                   discount=discount)
    dataset_demos = demonstration_dataset.map(transition)
    dataset = tf.data.experimental.sample_from_datasets(
        [dataset, dataset_demos],
        [1 - demonstration_ratio, demonstration_ratio])

    # Batch and prefetch.
    dataset = dataset.batch(batch_size, drop_remainder=True)
    dataset = dataset.prefetch(prefetch_size)

    # Use constant 0.05 epsilon greedy policy by default.
    if epsilon is None:
      epsilon = tf.Variable(0.05, trainable=False)
    policy_network = snt.Sequential([
        network,
        lambda q: trfl.epsilon_greedy(q, epsilon=epsilon).sample(),
    ])

    # Create a target network.
    target_network = copy.deepcopy(network)

    # Ensure that we create the variables before proceeding (maybe not needed).
    tf2_utils.create_variables(network, [environment_spec.observations])
    tf2_utils.create_variables(target_network, [environment_spec.observations])

    # Create the actor which defines how we take actions.
    actor = actors.FeedForwardActor(policy_network, adder)

    # The learner updates the parameters (and initializes them).
    learner = dqn.DQNLearner(
        network=network,
        target_network=target_network,
        discount=discount,
        importance_sampling_exponent=importance_sampling_exponent,
        learning_rate=learning_rate,
        target_update_period=target_update_period,
        dataset=dataset,
        replay_client=replay_client)

    super().__init__(
        actor=actor,
        learner=learner,
        min_observations=max(batch_size, min_replay_size),
        observations_per_step=float(batch_size) / samples_per_insert)


def _n_step_transition_from_episode(observations: acme_types.NestedTensor,
                                    actions: tf.Tensor,
                                    rewards: tf.Tensor,
                                    discounts: tf.Tensor,
                                    n_step: int,
                                    discount: float):
  """Produce Reverb-like N-step transition from a full episode.

  Observations, actions, rewards and discounts have the same length. This
  function will ignore the first reward and discount and the last action.

  Args:
    observations: [L, ...] Tensor.
    actions: [L, ...] Tensor.
    rewards: [L] Tensor.
    discounts: [L] Tensor.
    n_step: number of steps to squash into a single transition.
    discount: discount to use for TD updates.

  Returns:
    (o_t, a_t, r_t, d_t, o_tp1) tuple.
  """

  max_index = tf.shape(rewards)[0] - 1
  first = tf.random.uniform(shape=(), minval=0, maxval=max_index - 1,
                            dtype=tf.int32)
  last = tf.minimum(first + n_step, max_index)

  o_t = tree.map_structure(operator.itemgetter(first), observations)
  a_t = tree.map_structure(operator.itemgetter(first), actions)
  o_tp1 = tree.map_structure(operator.itemgetter(last), observations)

  # 0, 1, ..., n-1.
  discount_range = tf.cast(tf.range(last - first), tf.float32)
  # 1, g, ..., g^{n-1}.
  additional_discounts = tf.pow(discount, discount_range)
  # 1, d_t, d_t * d_{t+1}, ..., d_t * ... * d_{t+n-2}.
  discounts = tf.concat([[1.], tf.math.cumprod(discounts[first:last-1])], 0)
  # 1, g * d_t, ..., g^{n-1} * d_t * ... * d_{t+n-2}.
  discounts *= additional_discounts
  # r_t + g * d_t * r_{t+1} + ... + g^{n-1} * d_t * ... * d_{t+n-2} * r_{t+n-1}
  # We have to shift rewards by one so last=max_index corresponds to transitions
  # that include the last reward.
  r_t = tf.reduce_sum(rewards[first+1:last+1] * discounts)

  # g^{n-1} * d_{t} * ... * d_{t+n-1}.
  d_t = discounts[-1]

  key = tf.constant(0, tf.uint64)
  probability = tf.constant(1.0, tf.float64)
  table_size = tf.constant(1, tf.int64)
  priority = tf.constant(1.0, tf.float64)
  info = reverb.SampleInfo(
      key=key,
      probability=probability,
      table_size=table_size,
      priority=priority)
  return reverb.ReplaySample(
      info=info, data=acme_types.Transition(o_t, a_t, r_t, d_t, o_tp1))



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementations of a AIL agent."""

from acme.agents.jax.ail import losses
from acme.agents.jax.ail import rewards
from acme.agents.jax.ail.agents import AIL
from acme.agents.jax.ail.agents import DistributedAIL
from acme.agents.jax.ail.builder import AILBuilder
from acme.agents.jax.ail.config import AILConfig
from acme.agents.jax.ail.dac_agents import DAC
from acme.agents.jax.ail.dac_agents import DACConfig
from acme.agents.jax.ail.dac_agents import DistributedDAC
from acme.agents.jax.ail.gail_agents import DistributedGAIL
from acme.agents.jax.ail.gail_agents import GAIL
from acme.agents.jax.ail.gail_agents import GAILConfig
from acme.agents.jax.ail.learning import AILLearner
from acme.agents.jax.ail.networks import AILNetworks
from acme.agents.jax.ail.networks import AIRLModule
from acme.agents.jax.ail.networks import compute_ail_reward
from acme.agents.jax.ail.networks import DiscriminatorMLP
from acme.agents.jax.ail.networks import DiscriminatorModule
from acme.agents.jax.ail.networks import make_discriminator



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""PPO Builder."""
from typing import Callable, Iterator, List, Optional

from acme import adders
from acme import core
from acme import specs
from acme.adders import reverb as adders_reverb
from acme.agents.jax import actor_core as actor_core_lib
from acme.agents.jax import actors
from acme.agents.jax import builders
from acme.agents.jax.ppo import config as ppo_config
from acme.agents.jax.ppo import learning
from acme.agents.jax.ppo import networks as ppo_networks
from acme.jax import networks as networks_lib
from acme.jax import variable_utils
from acme.utils import counting
from acme.utils import loggers
import numpy as np
import optax
import reverb


class PPOBuilder(builders.ActorLearnerBuilder):
  """PPO Builder."""

  def __init__(
      self,
      config: ppo_config.PPOConfig,
      logger_fn: Callable[[], loggers.Logger] = lambda: None,
  ):
    """Creates PPO builder."""
    self._config = config
    self._logger_fn = logger_fn

    # An extra step is used for bootstrapping when computing advantages.
    self._sequence_length = config.unroll_length + 1

  def make_replay_tables(
      self, environment_spec: specs.EnvironmentSpec) -> List[reverb.Table]:
    """Creates reverb tables for the algorithm."""
    extra_spec = {
        'log_prob': np.ones(shape=(), dtype=np.float32),
    }
    signature = adders_reverb.SequenceAdder.signature(
        environment_spec, extra_spec, sequence_length=self._sequence_length)
    # We set a huge size of queue, even though it is not really necessary since
    # a custom update code in LocalLayout will make sure that we process data
    # as soon as it is available. So the size here does not really matter, as
    # long as it is bigger than or equal to the batch size.
    # CAVEAT: we cannot make it equal to batch_size or batch_size + 1, the value
    # has to be larger for few reasons:
    #  - during the last step in an episode we might write 2 sequences to reverb
    #    at once (that's how SequenceAdder works)
    #  - reverb does addition/sampling in multiple threads, so data might be
    #    added not when you expect it to be. so we need some extra buffer size
    #    in order not to get deadlocks.
    # TODO(raveman): This (large queue size) will not work for distributed setup
    # if learner is slower than actors (the queue will fill up and the learner
    # will be offpolicy). Find a better solution.
    is_single_machine = True
    extra_capacity_to_avoid_single_machine_deadlocks = 1000
    max_size = self._config.batch_size + (
        extra_capacity_to_avoid_single_machine_deadlocks
        if is_single_machine else 0)
    return [
        reverb.Table.queue(
            name=self._config.replay_table_name,
            max_size=max_size,
            signature=signature)
    ]

  def make_dataset_iterator(
      self, replay_client: reverb.Client) -> Iterator[reverb.ReplaySample]:
    """Creates a dataset."""
    # We don't use datasets.make_reverb_dataset() here to avoid interleaving
    # and prefetching, that doesn't work well with can_sample() check on update.
    # NOTE: Value for max_in_flight_samples_per_worker comes from a
    # recommendation here: https://git.io/JYzXB
    # TODO(bshahr): use TrajectoryDataset instead when the adders are updated.
    dataset = reverb.TrajectoryDataset.from_table_signature(
        server_address=replay_client.server_address,
        table=self._config.replay_table_name,
        max_in_flight_samples_per_worker=2 * self._config.batch_size)
    # Add batch dimension.
    dataset = dataset.batch(self._config.batch_size, drop_remainder=True)
    return dataset.as_numpy_iterator()

  def make_adder(self, replay_client: reverb.Client) -> adders.Adder:
    """Creates an adder which handles observations."""
    # Note that the last transition in the sequence is used for bootstrapping
    # only and is ignored otherwise. So we need to make sure that sequences
    # overlap on one transition, thus "-1" in the period length computation.
    return adders_reverb.SequenceAdder(
        client=replay_client,
        priority_fns={self._config.replay_table_name: None},
        period=self._sequence_length - 1,
        sequence_length=self._sequence_length,
        pad_end_of_episode=False,
        break_end_of_episode=False,
    )

  def make_learner(
      self,
      random_key: networks_lib.PRNGKey,
      networks: ppo_networks.PPONetworks,
      dataset: Iterator[reverb.ReplaySample],
      replay_client: Optional[reverb.Client] = None,
      counter: Optional[counting.Counter] = None,
  ) -> core.Learner:
    if callable(self._config.learning_rate):
      optimizer = optax.chain(
          optax.clip_by_global_norm(self._config.max_gradient_norm),
          optax.scale_by_adam(eps=self._config.adam_epsilon),
          optax.scale_by_schedule(self._config.learning_rate), optax.scale(-1))
    else:
      optimizer = optax.chain(
          optax.clip_by_global_norm(self._config.max_gradient_norm),
          optax.scale_by_adam(eps=self._config.adam_epsilon),
          optax.scale(-self._config.learning_rate))

    return learning.PPOLearner(
        ppo_networks=networks,
        iterator=dataset,
        discount=self._config.discount,
        entropy_cost=self._config.entropy_cost,
        value_cost=self._config.value_cost,
        max_abs_reward=self._config.max_abs_reward,
        ppo_clipping_epsilon=self._config.ppo_clipping_epsilon,
        clip_value=self._config.clip_value,
        gae_lambda=self._config.gae_lambda,
        counter=counter,
        random_key=random_key,
        optimizer=optimizer,
        num_epochs=self._config.num_epochs,
        num_minibatches=self._config.num_minibatches,
        logger=self._logger_fn(),
    )

  def make_actor(
      self,
      random_key: networks_lib.PRNGKey,
      policy_network,
      adder: Optional[adders.Adder] = None,
      variable_source: Optional[core.VariableSource] = None,
  ) -> core.Actor:
    assert variable_source is not None
    actor = actor_core_lib.batched_feed_forward_with_extras_to_actor_core(
        policy_network)
    variable_client = variable_utils.VariableClient(
        variable_source,
        'network',
        device='cpu',
        update_period=self._config.variable_update_period)
    return actors.GenericActor(
        actor, random_key, variable_client, adder, backend='cpu')



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the builder generator."""
from absl.testing import absltest
from acme import types
from acme.agents.jax.ail import builder
from jax import test_util as jtu
import numpy as np
import reverb

_REWARD = np.zeros((3,))


class BuilderTest(jtu.JaxTestCase):

  def test_weighted_generator(self):
    data0 = types.Transition(np.array([[1], [2], [3]]), (), _REWARD, (), ())
    it0 = iter([data0])

    data1 = types.Transition(np.array([[4], [5], [6]]), (), _REWARD, (), ())
    data2 = types.Transition(np.array([[7], [8], [9]]), (), _REWARD, (), ())
    it1 = iter([
        reverb.ReplaySample(info=reverb.SampleInfo((), (), (), ()), data=data1),
        reverb.ReplaySample(info=reverb.SampleInfo((), (), (), ()), data=data2)
    ])

    weighted_it = builder._generate_samples_with_demonstrations(
        it0, it1, policy_to_expert_data_ratio=2, batch_size=3)

    self.assertArraysEqual(
        next(weighted_it).data.observation, np.array([[1], [4], [5]]))
    self.assertArraysEqual(
        next(weighted_it).data.observation, np.array([[7], [8], [2]]))
    self.assertRaises(StopIteration, lambda: next(weighted_it))


if __name__ == '__main__':
  absltest.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the multi-objective MPO agent."""

from typing import Dict, Sequence, Tuple

from absl.testing import absltest
from absl.testing import parameterized
import acme
from acme import specs
from acme.agents.tf import mompo
from acme.testing import fakes
from acme.tf import networks
import numpy as np
import sonnet as snt
import tensorflow as tf


def make_networks(
    action_spec: specs.Array,
    num_critic_heads: int,
    policy_layer_sizes: Sequence[int] = (300, 200),
    critic_layer_sizes: Sequence[int] = (400, 300),
    num_layers_shared: int = 1,
    distributional_critic: bool = True,
    vmin: float = -150.,
    vmax: float = 150.,
    num_atoms: int = 51,
) -> Dict[str, snt.Module]:
  """Creates networks used by the agent."""

  num_dimensions = np.prod(action_spec.shape, dtype=int)

  policy_network = snt.Sequential([
      networks.LayerNormMLP(policy_layer_sizes, activate_final=True),
      networks.MultivariateNormalDiagHead(
          num_dimensions,
          tanh_mean=False,
          init_scale=0.69)
  ])

  if not distributional_critic:
    critic_layer_sizes = list(critic_layer_sizes) + [1]

  if not num_layers_shared:
    # No layers are shared
    critic_network_base = None
  else:
    critic_network_base = networks.LayerNormMLP(
        critic_layer_sizes[:num_layers_shared], activate_final=True)
  critic_network_heads = [
      snt.nets.MLP(critic_layer_sizes, activation=tf.nn.elu,
                   activate_final=False)
      for _ in range(num_critic_heads)]
  if distributional_critic:
    critic_network_heads = [
        snt.Sequential([
            c, networks.DiscreteValuedHead(vmin, vmax, num_atoms)
        ]) for c in critic_network_heads]
  # The multiplexer concatenates the (maybe transformed) observations/actions.
  critic_network = snt.Sequential([
      networks.CriticMultiplexer(
          critic_network=critic_network_base),
      networks.Multihead(network_heads=critic_network_heads),
  ])
  return {
      'policy': policy_network,
      'critic': critic_network,
  }


def compute_action_norm(target_pi_samples: tf.Tensor,
                        target_q_target_pi_samples: tf.Tensor) -> tf.Tensor:
  """Compute Q-values for the action norm objective from action samples."""
  del target_q_target_pi_samples
  action_norm = tf.norm(target_pi_samples, ord=2, axis=-1)
  return tf.stop_gradient(-1 * action_norm)


def task_reward_fn(observation: tf.Tensor,
                   action: tf.Tensor,
                   reward: tf.Tensor) -> tf.Tensor:
  del observation, action
  return tf.stop_gradient(reward)


def make_objectives() -> Tuple[
    Sequence[mompo.RewardObjective], Sequence[mompo.QValueObjective]]:
  """Define the multiple objectives for the policy to learn."""
  task_reward = mompo.RewardObjective(
      name='task',
      reward_fn=task_reward_fn)
  action_norm = mompo.QValueObjective(
      name='action_norm_q',
      qvalue_fn=compute_action_norm)
  return [task_reward], [action_norm]


class MOMPOTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('distributional_critic', True),
      ('vanilla_critic', False))
  def test_mompo(self, distributional_critic):
    # Create a fake environment to test with.
    environment = fakes.ContinuousEnvironment(episode_length=10)
    spec = specs.make_environment_spec(environment)

    # Create objectives.
    reward_objectives, qvalue_objectives = make_objectives()
    num_critic_heads = len(reward_objectives)

    # Create networks.
    agent_networks = make_networks(
        spec.actions, num_critic_heads=num_critic_heads,
        distributional_critic=distributional_critic)

    # Construct the agent.
    agent = mompo.MultiObjectiveMPO(
        reward_objectives,
        qvalue_objectives,
        spec,
        policy_network=agent_networks['policy'],
        critic_network=agent_networks['critic'],
        batch_size=10,
        samples_per_insert=2,
        min_replay_size=10)

    # Try running the environment loop. We have no assertions here because all
    # we care about is that the agent runs without raising any errors.
    loop = acme.EnvironmentLoop(environment, agent)
    loop.run(num_episodes=2)


if __name__ == '__main__':
  absltest.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""SAC learner implementation."""

import time
from typing import Any, Dict, Iterator, List, NamedTuple, Optional, Tuple

import acme
from acme import types
from acme.agents.jax.sac import networks as sac_networks
from acme.jax import networks as networks_lib
from acme.jax import utils
from acme.utils import counting
from acme.utils import loggers
import jax
import jax.numpy as jnp
import optax
import reverb


class TrainingState(NamedTuple):
  """Contains training state for the learner."""
  policy_optimizer_state: optax.OptState
  q_optimizer_state: optax.OptState
  policy_params: networks_lib.Params
  q_params: networks_lib.Params
  target_q_params: networks_lib.Params
  key: networks_lib.PRNGKey
  alpha_optimizer_state: Optional[optax.OptState] = None
  alpha_params: Optional[networks_lib.Params] = None


class SACLearner(acme.Learner):
  """SAC learner."""

  _state: TrainingState

  def __init__(
      self,
      networks: sac_networks.SACNetworks,
      rng: jnp.ndarray,
      iterator: Iterator[reverb.ReplaySample],
      policy_optimizer: optax.GradientTransformation,
      q_optimizer: optax.GradientTransformation,
      tau: float = 0.005,
      reward_scale: float = 1.0,
      discount: float = 0.99,
      entropy_coefficient: Optional[float] = None,
      target_entropy: float = 0,
      counter: Optional[counting.Counter] = None,
      logger: Optional[loggers.Logger] = None,
      num_sgd_steps_per_step: int = 1):
    """Initialize the SAC learner.

    Args:
      networks: SAC networks
      rng: a key for random number generation.
      iterator: an iterator over training data.
      policy_optimizer: the policy optimizer.
      q_optimizer: the Q-function optimizer.
      tau: target smoothing coefficient.
      reward_scale: reward scale.
      discount: discount to use for TD updates.
      entropy_coefficient: coefficient applied to the entropy bonus. If None, an
        adaptative coefficient will be used.
      target_entropy: Used to normalize entropy. Only used when
        entropy_coefficient is None.
      counter: counter object used to keep track of steps.
      logger: logger object to be used by learner.
      num_sgd_steps_per_step: number of sgd steps to perform per learner 'step'.
    """
    adaptive_entropy_coefficient = entropy_coefficient is None
    if adaptive_entropy_coefficient:
      # alpha is the temperature parameter that determines the relative
      # importance of the entropy term versus the reward.
      log_alpha = jnp.asarray(0., dtype=jnp.float32)
      alpha_optimizer = optax.adam(learning_rate=3e-4)
      alpha_optimizer_state = alpha_optimizer.init(log_alpha)
    else:
      if target_entropy:
        raise ValueError('target_entropy should not be set when '
                         'entropy_coefficient is provided')

    def alpha_loss(log_alpha: jnp.ndarray,
                   policy_params: networks_lib.Params,
                   transitions: types.Transition,
                   key: networks_lib.PRNGKey) -> jnp.ndarray:
      """Eq 18 from https://arxiv.org/pdf/1812.05905.pdf."""
      dist_params = networks.policy_network.apply(
          policy_params, transitions.observation)
      action = networks.sample(dist_params, key)
      log_prob = networks.log_prob(dist_params, action)
      alpha = jnp.exp(log_alpha)
      alpha_loss = alpha * jax.lax.stop_gradient(-log_prob - target_entropy)
      return jnp.mean(alpha_loss)

    def critic_loss(q_params: networks_lib.Params,
                    policy_params: networks_lib.Params,
                    target_q_params: networks_lib.Params,
                    alpha: jnp.ndarray,
                    transitions: types.Transition,
                    key: networks_lib.PRNGKey) -> jnp.ndarray:
      q_old_action = networks.q_network.apply(
          q_params, transitions.observation, transitions.action)
      next_dist_params = networks.policy_network.apply(
          policy_params, transitions.next_observation)
      next_action = networks.sample(next_dist_params, key)
      next_log_prob = networks.log_prob(next_dist_params, next_action)
      next_q = networks.q_network.apply(
          target_q_params, transitions.next_observation, next_action)
      next_v = jnp.min(next_q, axis=-1) - alpha * next_log_prob
      target_q = jax.lax.stop_gradient(transitions.reward * reward_scale +
                                       transitions.discount * discount * next_v)
      q_error = q_old_action - jnp.expand_dims(target_q, -1)
      q_loss = 0.5 * jnp.mean(jnp.square(q_error))
      return q_loss

    def actor_loss(policy_params: networks_lib.Params,
                   q_params: networks_lib.Params,
                   alpha: jnp.ndarray,
                   transitions: types.Transition,
                   key: networks_lib.PRNGKey) -> jnp.ndarray:
      dist_params = networks.policy_network.apply(
          policy_params, transitions.observation)
      action = networks.sample(dist_params, key)
      log_prob = networks.log_prob(dist_params, action)
      q_action = networks.q_network.apply(
          q_params, transitions.observation, action)
      min_q = jnp.min(q_action, axis=-1)
      actor_loss = alpha * log_prob - min_q
      return jnp.mean(actor_loss)

    alpha_grad = jax.value_and_grad(alpha_loss)
    critic_grad = jax.value_and_grad(critic_loss)
    actor_grad = jax.value_and_grad(actor_loss)

    def update_step(
        state: TrainingState,
        transitions: types.Transition,
    ) -> Tuple[TrainingState, Dict[str, jnp.ndarray]]:

      key, key_alpha, key_critic, key_actor = jax.random.split(state.key, 4)
      if adaptive_entropy_coefficient:
        alpha_loss, alpha_grads = alpha_grad(state.alpha_params,
                                             state.policy_params, transitions,
                                             key_alpha)
        alpha = jnp.exp(state.alpha_params)
      else:
        alpha = entropy_coefficient
      critic_loss, critic_grads = critic_grad(state.q_params,
                                              state.policy_params,
                                              state.target_q_params, alpha,
                                              transitions, key_critic)
      actor_loss, actor_grads = actor_grad(state.policy_params, state.q_params,
                                           alpha, transitions, key_actor)

      # Apply policy gradients
      actor_update, policy_optimizer_state = policy_optimizer.update(
          actor_grads, state.policy_optimizer_state)
      policy_params = optax.apply_updates(state.policy_params, actor_update)

      # Apply critic gradients
      critic_update, q_optimizer_state = q_optimizer.update(
          critic_grads, state.q_optimizer_state)
      q_params = optax.apply_updates(state.q_params, critic_update)

      new_target_q_params = jax.tree_multimap(
          lambda x, y: x * (1 - tau) + y * tau, state.target_q_params, q_params)

      metrics = {
          'critic_loss': critic_loss,
          'actor_loss': actor_loss,
      }

      new_state = TrainingState(
          policy_optimizer_state=policy_optimizer_state,
          q_optimizer_state=q_optimizer_state,
          policy_params=policy_params,
          q_params=q_params,
          target_q_params=new_target_q_params,
          key=key,
      )
      if adaptive_entropy_coefficient:
        # Apply alpha gradients
        alpha_update, alpha_optimizer_state = alpha_optimizer.update(
            alpha_grads, state.alpha_optimizer_state)
        alpha_params = optax.apply_updates(state.alpha_params, alpha_update)
        metrics.update({
            'alpha_loss': alpha_loss,
            'alpha': jnp.exp(alpha_params),
        })
        new_state = new_state._replace(
            alpha_optimizer_state=alpha_optimizer_state,
            alpha_params=alpha_params)

      metrics['observations_mean'] = jnp.mean(
          utils.batch_concat(
              jax.tree_map(lambda x: jnp.abs(jnp.mean(x, axis=0)),
                           transitions.observation)))
      metrics['observations_std'] = jnp.mean(
          utils.batch_concat(
              jax.tree_map(lambda x: jnp.std(x, axis=0),
                           transitions.observation)))
      metrics['next_observations_mean'] = jnp.mean(
          utils.batch_concat(
              jax.tree_map(lambda x: jnp.abs(jnp.mean(x, axis=0)),
                           transitions.next_observation)))
      metrics['next_observations_std'] = jnp.mean(
          utils.batch_concat(
              jax.tree_map(lambda x: jnp.std(x, axis=0),
                           transitions.next_observation)))

      return new_state, metrics

    # General learner book-keeping and loggers.
    self._counter = counter or counting.Counter()
    self._logger = logger or loggers.make_default_logger(
        'learner', asynchronous=True, serialize_fn=utils.fetch_devicearray)

    # Iterator on demonstration transitions.
    self._iterator = iterator

    update_step = utils.process_multiple_batches(update_step,
                                                 num_sgd_steps_per_step)
    # Use the JIT compiler.
    self._update_step = jax.jit(update_step)

    def make_initial_state(key: networks_lib.PRNGKey) -> TrainingState:
      """Initialises the training state (parameters and optimiser state)."""
      key_policy, key_q, key = jax.random.split(key, 3)

      policy_params = networks.policy_network.init(key_policy)
      policy_optimizer_state = policy_optimizer.init(policy_params)

      q_params = networks.q_network.init(key_q)
      q_optimizer_state = q_optimizer.init(q_params)

      state = TrainingState(
          policy_optimizer_state=policy_optimizer_state,
          q_optimizer_state=q_optimizer_state,
          policy_params=policy_params,
          q_params=q_params,
          target_q_params=q_params,
          key=key)

      if adaptive_entropy_coefficient:
        state = state._replace(alpha_optimizer_state=alpha_optimizer_state,
                               alpha_params=log_alpha)
      return state

    # Create initial state.
    self._state = make_initial_state(rng)

    # Do not record timestamps until after the first learning step is done.
    # This is to avoid including the time it takes for actors to come online and
    # fill the replay buffer.
    self._timestamp = None

  def step(self):
    sample = next(self._iterator)
    transitions = types.Transition(*sample.data)

    self._state, metrics = self._update_step(self._state, transitions)

    # Compute elapsed time.
    timestamp = time.time()
    elapsed_time = timestamp - self._timestamp if self._timestamp else 0
    self._timestamp = timestamp

    # Increment counts and record the current time
    counts = self._counter.increment(steps=1, walltime=elapsed_time)

    # Attempts to write the logs.
    self._logger.write({**metrics, **counts})

  def get_variables(self, names: List[str]) -> List[Any]:
    variables = {
        'policy': self._state.policy_params,
    }
    return [variables[name] for name in names]

  def save(self) -> TrainingState:
    return self._state

  def restore(self, state: TrainingState):
    self._state = state



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Offline losses used in variants of BC."""
from typing import Any, Callable, Optional

from acme import types
from acme.jax import networks as networks_lib
import jax
import jax.numpy as jnp


ModelOutput = Any
SampleFn = Callable[[ModelOutput, networks_lib.PRNGKey], networks_lib.Action]
LogProb = jnp.ndarray
LogProbFn = Callable[[ModelOutput, networks_lib.Action], LogProb]
Loss = Callable[[
    Callable[..., networks_lib.NetworkOutput], networks_lib.Params, networks_lib
    .PRNGKey, types.Transition
], jnp.ndarray]


def mse(sample_fn: SampleFn) -> Loss:
  """Mean Squared Error loss.

  Args:
    sample_fn: a method that samples an action.
  Returns:
    The loss.
  """

  def loss(apply_fn: Callable[..., networks_lib.NetworkOutput],
           params: networks_lib.Params, key: jnp.ndarray,
           transitions: types.Transition) -> jnp.ndarray:
    key, key_dropout = jax.random.split(key)
    dist_params = apply_fn(
        params, transitions.observation, is_training=True, key=key_dropout)
    action = sample_fn(dist_params, key)
    return jnp.mean(jnp.square(action - transitions.action))

  return loss


def logp(logp_fn: LogProbFn) -> Loss:
  """Log probability loss.

  Args:
    logp_fn: a method that returns the log probability of an action.
  Returns:
    The loss.
  """

  def loss(apply_fn: Callable[..., networks_lib.NetworkOutput],
           params: networks_lib.Params, key: jnp.ndarray,
           transitions: types.Transition) -> jnp.ndarray:
    logits = apply_fn(
        params, transitions.observation, is_training=True, key=key)
    logp_action = logp_fn(logits, transitions.action)
    return -jnp.mean(logp_action)

  return loss


def peerbc(base_loss_fn: Loss, zeta: float) -> Loss:
  """Peer-BC loss from https://arxiv.org/pdf/2010.01748.pdf.

  Args:
    base_loss_fn: the base loss to add RCAL on top of.
    zeta: the weight of the regularization.
  Returns:
    The loss.
  """

  def loss(apply_fn: Callable[..., networks_lib.NetworkOutput],
           params: networks_lib.Params, key: jnp.ndarray,
           transitions: types.Transition) -> jnp.ndarray:
    key_perm, key_bc_loss, key_permuted_loss = jax.random.split(key, 3)

    permutation_keys = jax.random.split(key_perm, transitions.action.shape[0])
    permuted_actions = jax.vmap(
        jax.random.permutation, in_axes=(0, 0))(permutation_keys,
                                                transitions.action)
    permuted_transition = transitions._replace(action=permuted_actions)
    bc_loss = base_loss_fn(apply_fn, params, key_bc_loss, transitions)
    permuted_loss = base_loss_fn(apply_fn, params, key_permuted_loss,
                                 permuted_transition)
    return bc_loss - zeta * permuted_loss
  return loss


def rcal(base_loss_fn: Loss,
         discount: float,
         alpha: float,
         num_bins: Optional[int] = None) -> Loss:
  """https://www.cristal.univ-lille.fr/~pietquin/pdf/AAMAS_2014_BPMGOP.pdf.

  Args:
    base_loss_fn: the base loss to add RCAL on top of.
    discount: the gamma discount used in RCAL.
    alpha: the regularization parameter.
    num_bins: how many bins were used for discretization. If None the
      environment was originally discrete already.
  Returns:
    The loss function.
  """

  def loss(apply_fn: Callable[..., networks_lib.NetworkOutput],
           params: networks_lib.Params, key: jnp.ndarray,
           transitions: types.Transition) -> jnp.ndarray:

    def logits_fn(key, observations, actions=None):
      logits = apply_fn(params, observations, key=key, is_training=True)
      if num_bins:
        logits = jnp.reshape(logits, list(logits.shape[:-1]) + [-1, num_bins])
      if actions is None:
        actions = jnp.argmax(logits, axis=-1)
      logits_actions = jnp.sum(
          jax.nn.one_hot(actions, logits.shape[-1]) * logits, axis=-1)
      return logits_actions

    key, key1, key2 = jax.random.split(key, 3)

    logits_a_tm1 = logits_fn(key1, transitions.observation, transitions.action)
    logits_a_t = logits_fn(key2, transitions.next_observation)

    # RCAL, by making a parallel between the logits of BC and Q-values,
    # defines a regularization loss that encourages the implicit reward
    # (inferred by inversing the Bellman Equation) to be sparse.
    # NOTE: In case of discretized envs jnp.mean goes over batch and num_bins
    # dimensions.
    regularization_loss = jnp.mean(
        jnp.abs(logits_a_tm1 - discount * logits_a_t)
        )

    loss = base_loss_fn(apply_fn, params, key, transitions)
    return loss + alpha * regularization_loss
  return loss



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wrapper that implements concatenation of observation fields."""

from typing import Sequence, Optional

from acme import types
from acme.wrappers import base
import dm_env
import numpy as np
import tree


def _concat(values: types.NestedArray) -> np.ndarray:
  """Concatenates the leaves of `values` along the leading dimension.

  Treats scalars as 1d arrays and expects that the shapes of all leaves are
  the same except for the leading dimension.

  Args:
    values: the nested arrays to concatenate.
  Returns:
    The concatenated array.
  """
  leaves = list(map(np.atleast_1d, tree.flatten(values)))
  return np.concatenate(leaves)


def _zeros_like(nest, dtype=None):
  """Generate a nested NumPy array according to spec."""
  return tree.map_structure(lambda x: np.zeros(x.shape, dtype or x.dtype), nest)


class ConcatObservationWrapper(base.EnvironmentWrapper):
  """Wrapper that concatenates observation fields.

  It takes an environment with nested observations and concatenates the fields
  in a single tensor. The orginial fields should be 1-dimensional.
  Observation fields that are not in name_filter are dropped.
  """

  def __init__(self, environment: dm_env.Environment,
               name_filter: Optional[Sequence[str]] = None):
    """Initializes a new ConcatObservationWrapper.

    Args:
      environment: Environment to wrap.
      name_filter: Sequence of observation names to keep. None keeps them all.
    """
    super().__init__(environment)
    observation_spec = environment.observation_spec()
    if name_filter is None:
      name_filter = list(observation_spec.keys())
    self._obs_names = [x for x in name_filter if x in observation_spec.keys()]

    dummy_obs = _zeros_like(observation_spec)
    dummy_obs = self._convert_observation(dummy_obs)
    self._observation_spec = dm_env.specs.BoundedArray(
        shape=dummy_obs.shape,
        dtype=dummy_obs.dtype,
        minimum=-np.inf,
        maximum=np.inf,
        name='state')

  def _convert_observation(self, observation):
    obs = {k: observation[k] for k in self._obs_names}
    return _concat(obs)

  def step(self, action) -> dm_env.TimeStep:
    timestep = self._environment.step(action)
    return timestep._replace(
        observation=self._convert_observation(timestep.observation))

  def reset(self) -> dm_env.TimeStep:
    timestep = self._environment.reset()
    return timestep._replace(
        observation=self._convert_observation(timestep.observation))

  def observation_spec(self) -> types.NestedSpec:
    return self._observation_spec



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Integration test for the distributed agent."""

from absl.testing import absltest
import acme
from acme import specs
from acme.agents.tf import d4pg
from acme.testing import fakes
from acme.tf import networks
from acme.tf import utils as tf2_utils
import launchpad as lp
import numpy as np
import sonnet as snt


def make_networks(action_spec: specs.BoundedArray):
  """Simple networks for testing.."""

  num_dimensions = np.prod(action_spec.shape, dtype=int)

  policy_network = snt.Sequential([
      networks.LayerNormMLP([50], activate_final=True),
      networks.NearZeroInitializedLinear(num_dimensions),
      networks.TanhToSpec(action_spec)
  ])
  # The multiplexer concatenates the (maybe transformed) observations/actions.
  critic_network = snt.Sequential([
      networks.CriticMultiplexer(
          critic_network=networks.LayerNormMLP(
              [50], activate_final=True)),
      networks.DiscreteValuedHead(-1., 1., 10)
  ])

  return {
      'policy': policy_network,
      'critic': critic_network,
      'observation': tf2_utils.batch_concat,
  }


class DistributedAgentTest(absltest.TestCase):
  """Simple integration/smoke test for the distributed agent."""

  def test_control_suite(self):
    """Tests that the agent can run on the control suite without crashing."""

    agent = d4pg.DistributedD4PG(
        environment_factory=lambda x: fakes.ContinuousEnvironment(bounded=True),
        network_factory=make_networks,
        num_actors=2,
        batch_size=32,
        min_replay_size=32,
        max_replay_size=1000,
    )
    program = agent.build()

    (learner_node,) = program.groups['learner']
    learner_node.disable_run()

    lp.launch(program, launch_type='test_mt')

    learner: acme.Learner = learner_node.create_handle().dereference()

    for _ in range(5):
      learner.step()


if __name__ == '__main__':
  absltest.main()
